"""Tests for the vanilla NN."""

import numpy as np

from nsquared.vanilla_nn import row_row, col_col

# initialize the number of rows and columns
ROWS = 10
COLS = 10
# initialize the NN imputer
imputer = row_row(
    distance_threshold=1,
    is_percentile=False,
)


def test_constant_imputation_0() -> None:
    """Sanity Check: Given a matrix with constant values,
    the imputation should be the same constant value.
    """
    data = np.zeros((ROWS, COLS), dtype=float)
    mask = np.ones((ROWS, COLS), dtype=bool)

    for r, c in np.ndindex(data.shape):
        estimated_value = imputer.impute(
            row=r,
            column=c,
            data_array=data,
            mask_array=mask,
        )
        assert estimated_value == 0


def test_constant_imputation_0_5() -> None:
    """Sanity Check: Given a matrix with constant values,
    the imputation should be the same constant value.
    """
    data = np.ones((ROWS, COLS), dtype=float) * 0.5
    mask = np.ones((ROWS, COLS), dtype=bool)

    for r, c in np.ndindex(data.shape):
        estimated_value = imputer.impute(
            row=r,
            column=c,
            data_array=data,
            mask_array=mask,
        )
        assert estimated_value == 0.5


def test_constant_imputation_1() -> None:
    """Sanity Check: Given a matrix with constant values,
    the imputation should be the same constant value.
    """
    data = np.ones((ROWS, COLS), dtype=float)
    mask = np.ones((ROWS, COLS), dtype=bool)

    for r, c in np.ndindex(data.shape):
        estimated_value = imputer.impute(
            row=r,
            column=c,
            data_array=data,
            mask_array=mask,
        )
        assert estimated_value == 1


def test_half_ones_half_zeros_observed_1() -> None:
    """Given a data matrix with the first half of the columns being ones
    and the second half of the columns being zeros, but only the first
    half is observed, the imputed value should be one for the first half
    and zero for the second half.
    """
    data = np.ones((ROWS, COLS), dtype=float)
    data[:, COLS // 2 :] = 0
    mask = np.ones((ROWS, COLS), dtype=bool)
    mask[:, COLS // 2 :] = 0

    for r in range(ROWS):
        for c in range(COLS):
            if c < COLS // 2:
                estimated_value = imputer.impute(
                    row=r,
                    column=c,
                    data_array=data,
                    mask_array=mask,
                )
                assert estimated_value == 1
            else:
                estimated_value = imputer.impute(
                    row=r,
                    column=c,
                    data_array=data,
                    mask_array=mask,
                )
                assert np.isnan(estimated_value)


def test_half_ones_half_zeros_observed_2() -> None:
    """Given a data matrix with the first half of the columns being ones
    and the second half of the columns being zeros, but only the second
    half is observed, the imputed value should be one for the first half
    and zero for the second half.
    """
    data = np.ones((ROWS, COLS), dtype=float)
    data[:, COLS // 2 :] = 0
    mask = np.ones((ROWS, COLS), dtype=bool)
    mask[:, : COLS // 2] = 0

    for r in range(ROWS):
        for c in range(COLS):
            if c < COLS // 2:
                estimated_value = imputer.impute(
                    row=r,
                    column=c,
                    data_array=data,
                    mask_array=mask,
                )
                assert np.isnan(estimated_value)
            else:
                estimated_value = imputer.impute(
                    row=r,
                    column=c,
                    data_array=data,
                    mask_array=mask,
                )
                assert estimated_value == 0


def test_example_1() -> None:
    """Test the entry at (1, 2)."""
    DATA = np.array(
        [
            [1, 1, 1, 1],
            [1, 1, 0, 1],
            [1, 1, 1, 1],
            [1, 1, 0, 1],
        ],
        dtype=float,
    )
    MASK = np.array(
        [
            [1, 1, 1, 1],
            [1, 1, 1, 1],
            [1, 1, 1, 1],
            [1, 1, 1, 1],
        ],
        dtype=bool,
    )
    imputer = row_row(
        distance_threshold=0,
    )
    estimated_value = imputer.impute(
        row=1,
        column=2,
        data_array=DATA,
        mask_array=MASK,
    )
    assert np.isclose(estimated_value, 0.5)


def test_example_2() -> None:
    """Test the entry at (1, 2) using row-row imputation."""
    DATA = np.array(
        [
            [1, 1, 1, 1],
            [1, 1, 0, 1],
            [1, 1, 1, 1],
            [1, 1, 0, 1],
        ],
        dtype=float,
    )
    MASK = np.array(
        [
            [1, 1, 1, 1],
            [1, 1, 0, 1],
            [1, 1, 1, 1],
            [1, 1, 1, 1],
        ],
        dtype=bool,
    )
    imputer = row_row(
        distance_threshold=0,
    )
    estimated_value = imputer.impute(
        row=1,
        column=2,
        data_array=DATA,
        mask_array=MASK,
    )
    assert np.isclose(estimated_value, 2 / 3)


def test_example_3() -> None:
    """Test the entry at (1, 2) using col-col imputation.

    NOTE: we simply transpose the DATA and MASK matrices
    from test_example_2.
    """
    DATA = np.array(
        [
            [1, 1, 1, 1],
            [1, 1, 1, 1],
            [1, 0, 1, 0],
            [1, 1, 1, 1],
        ],
        dtype=float,
    )
    MASK = np.array(
        [
            [1, 1, 1, 1],
            [1, 1, 1, 1],
            [1, 0, 1, 1],
            [1, 1, 1, 1],
        ],
        dtype=bool,
    )
    imputer = col_col(
        distance_threshold=0,
    )
    estimated_value = imputer.impute(
        row=2,
        column=1,
        data_array=DATA,
        mask_array=MASK,
    )
    assert np.isclose(estimated_value, 2 / 3)



# from numpy._core.numerictypes import float
from .nnimputer import EstimationMethod, DataType
import numpy.typing as npt
import numpy as np
from typing import Union, Tuple, Any
import logging
import warnings
from typing import Optional
from .data_types import Scalar
from tqdm import tqdm

logger = logging.getLogger(__name__)


class RowRowEstimator(EstimationMethod):
    """Estimate the missing value using row-row nearest neighbors."""

    def __init__(self, is_percentile: bool = True):
        """Initialize the row-row estimator.

        Args:
            is_percentile (bool): Whether to use percentile-based threshold. Defaults to True.

        """
        super().__init__(is_percentile)
        self.row_distances = dict()

    def __str__(self):
        return "RowRowEstimator"

    def impute(
        self,
        row: int,
        column: int,
        data_array: npt.NDArray,
        mask_array: npt.NDArray,
        distance_threshold: Union[float, Tuple[float, float]],
        data_type: DataType,
        allow_self_neighbor: bool = False,
        **kwargs: Any,
    ) -> npt.NDArray:
        """Impute the missing value at the given row and column.

        Args:
            row (int): Row index
            column (int): Column index
            data_array (npt.NDArray): Data matrix
            mask_array (npt.NDArray): Mask matrix
            distance_threshold (float): Distance threshold for nearest neighbors
            data_type (DataType): Data type to use (e.g. scalars, distributions)
            allow_self_neighbor (bool): Whether to allow self-neighbor. Defaults to False.
            **kwargs (Any): Additional keyword arguments

        Returns:
            npt.NDArray: Imputed value

        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            # Calculate distances between rows
            self._calculate_distances(row, column, data_array, mask_array, data_type)
            all_dists = np.copy(self.row_distances[row])
            # Exclude the target column
            if not allow_self_neighbor:
                all_dists[:, column] = np.nan
            row_dists = np.nanmean(all_dists, axis=1)
            if self.is_percentile:
                # NOTE: we assume eta_row and eta_col are in [0, 1] in this case
                quantile_row_dists = row_dists[
                    ~np.isnan(row_dists) & (row_dists != np.inf)
                ]
                if quantile_row_dists.size == 0:
                    # No valid distances, return np.nan
                    eta_row = np.nan
                else:
                    eta_row = np.quantile(quantile_row_dists, distance_threshold)
            else:
                eta_row = distance_threshold

            # Find the nearest neighbors indexes
            nearest_neighbors = np.where(row_dists <= eta_row)[0]

        # NOTE: this code block will never be called since the target row
        # is always a nearest neighbor
        if len(nearest_neighbors) == 0:
            # return np.array(np.nan)
            # NOTE: implement the base case described by Eq. 11 in
            # "Counterfactual Inference for Sequential Experiments".
            if mask_array[row, column]:
                # return the observed outcome
                return data_array[row, column]
            else:
                # return the average of all observed outcomes corresponding
                # to treatment 1 at time t.
                return data_type.average(data_array[:, column])

        # Calculate the average of the nearest neighbors
        nearest_neighbors_data = data_array[nearest_neighbors, column]
        nearest_neighbors_mask = mask_array[nearest_neighbors, column]
        nearest_neighbors_data = nearest_neighbors_data[nearest_neighbors_mask == 1]
        return data_type.average(nearest_neighbors_data)

    def _calculate_distances(
        self,
        row: int,
        col: int,
        data_array: npt.NDArray,
        mask_array: npt.NDArray,
        data_type: DataType,
    ) -> None:
        """Sets the distances for the imputer.
        Sets the distances as a class attribute, so returns nothing.

        Args:
            row (int): Row index
            col (int): Column index
            data_array (npt.NDArray): Data matrix
            mask_array (npt.NDArray): Mask matrix
            data_type (DataType): Data type to use (e.g. scalars, distributions)

        """
        data_shape = data_array.shape
        n_rows = data_shape[0]
        n_cols = data_shape[1]

        if row in self.row_distances:
            return

        # Calculate distances between rows
        row_dists = np.zeros((n_rows, n_cols))
        # Scalar optimization with vectorized operations instead of loops
        if isinstance(data_type, Scalar):
            # Determine overlap columns for any pairwise rows
            overlap_columns_mask = np.logical_and(
                np.tile(mask_array[row], (n_rows, 1)), mask_array
            )
            row_big_matrix = np.tile(data_array[row], (n_rows, 1))
            row_dists = np.power(data_array - row_big_matrix, 2).astype(np.float64)
            # We need the row dists as a float matrix to use np.nanmean
            row_dists[~overlap_columns_mask] = np.nan
            self.row_distances[row] = row_dists
            return

        for i in range(n_rows):
            # Get columns observed in both row i and row
            overlap_columns = np.logical_and(mask_array[row], mask_array[i])

            if not np.any(overlap_columns):
                row_dists[i, :] = np.nan
                continue

            # Calculate distance between rows
            for j in range(n_cols):
                if not overlap_columns[j]:  # Skip missing values and the target column
                    row_dists[i, j] = np.nan
                else:
                    # if data_array[row, j].dtype == np.float64:
                    #     print (f"Row {row}, Column {j} is NaN")
                    #     exit()
                    # if isinstance(data_array[i, j], np.float64):
                    #     print (f"Row {i}, Column {j} is NaN")
                    #     exit()
                    row_dists[i, j] = data_type.distance(
                        data_array[row, j], data_array[i, j]
                    )
        self.row_distances[row] = row_dists


class ColColEstimator(EstimationMethod):
    """Estimate the missing value using column-column nearest neighbors."""

    def __init__(self, is_percentile: bool = True):
        """Initialize the column-column estimator.

        Args:
            is_percentile (bool): Whether to use percentile-based threshold. Defaults to True.

        """
        self.estimator = RowRowEstimator(is_percentile=is_percentile)
        # use the same logic as RowRowEstimator but transposed
        # save the distances

    def __str__(self):
        return "ColColEstimator"

    def impute(
        self,
        row: int,
        column: int,
        data_array: npt.NDArray,
        mask_array: npt.NDArray,
        distance_threshold: Union[float, Tuple[float, float]],
        data_type: DataType,
        allow_self_neighbor: bool = False,
        **kwargs: Any,
    ) -> npt.NDArray:
        """Impute the missing value at the given row and column.

        Args:
            row (int): Row index
            column (int): Column index
            data_array (npt.NDArray): Data matrix
            mask_array (npt.NDArray): Mask matrix
            distance_threshold (float): Distance threshold for nearest neighbors
            data_type (DataType): Data type to use (e.g. scalars, distributions)
            allow_self_neighbor (bool): Whether to allow self-neighbor. Defaults to False.
            **kwargs (Any): Additional keyword arguments

        Returns:
            npt.NDArray: Imputed value

        """
        data_transposed = np.swapaxes(data_array, 0, 1)
        mask_transposed = np.swapaxes(mask_array, 0, 1)

        return self.estimator.impute(
            column,
            row,
            data_transposed,
            mask_transposed,
            distance_threshold,
            data_type,
            allow_self_neighbor,
        )

    def _calculate_distances(
        self,
        row: int,
        col: int,
        data_array: npt.NDArray,
        mask_array: npt.NDArray,
        data_type: DataType,
    ) -> None:
        """Sets the distances for the imputer.
        Sets the distances as a class attribute, so returns nothing.

        Args:
            row (int): Row index
            col (int): Column index
            data_array (npt.NDArray): Data matrix
            mask_array (npt.NDArray): Mask matrix
            data_type (DataType): Data type to use (e.g. scalars, distributions)

        """
        # We use the RowRow Estimator to calculate the distances
        pass


class DREstimator(EstimationMethod):
    """Estimate the missing entry using doubly robust nearest neighbors."""

    def __init__(self, is_percentile: bool = True):
        """Initialize the doubly robust estimator.

        Args:
            is_percentile (bool): Whether to use percentile-based threshold. Defaults to True.

        """
        super().__init__(is_percentile)
        self.row_distances = dict()
        self.col_distances = dict()

    def impute(
        self,
        row: int,
        column: int,
        data_array: npt.NDArray,
        mask_array: npt.NDArray,
        distance_threshold: Union[float, Tuple[float, float]],
        data_type: DataType,
        allow_self_neighbor: bool = False,
        **kwargs: Any,
    ) -> npt.NDArray:
        """Impute the missing value at the given row and column using doubly robust method.

        Args:
        ----
            row (int): Row index
            column (int): Column index
            data_array (npt.NDArray): Data matrix
            mask_array (npt.NDArray): Mask matrix
            distance_threshold (float or Tuple[float, float]): Distance threshold for nearest neighbors
            or a tuple of (row_threshold, col_threshold) for row and column respectively.
            data_type (DataType): Data type to use (e.g. scalars, distributions)
            allow_self_neighbor (bool): Whether to allow self-neighbor. Defaults to False.
            **kwargs (Any): Additional keyword arguments

        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            self._calculate_distances(row, column, data_array, mask_array, data_type)
            all_dists = np.copy(self.row_distances[row])
            # Exclude the target column
            if not allow_self_neighbor:
                all_dists[:, column] = np.nan
            row_dists = np.nanmean(all_dists, axis=1)
            if not mask_array[row, column] and not allow_self_neighbor:
                row_dists[row] = np.inf  # Exclude the target row
            # Exclude the target row
            all_dists = np.copy(self.col_distances[column])
            if not allow_self_neighbor:
                all_dists[row, :] = np.nan
            col_dists = np.nanmean(all_dists, axis=0)
            if not mask_array[row, column] and not allow_self_neighbor:
                col_dists[column] = np.inf  # Exclude the target col

        if isinstance(distance_threshold, tuple):
            eta_row = distance_threshold[0]
            eta_col = distance_threshold[1]
        else:
            eta_row = distance_threshold
            eta_col = distance_threshold
        if self.is_percentile:
            # NOTE: we assume eta_row and eta_col are in [0, 1] in this case
            quantile_row_dists = row_dists[~np.isnan(row_dists) & (row_dists != np.inf)]
            quantile_col_dists = col_dists[~np.isnan(col_dists) & (col_dists != np.inf)]
            if quantile_row_dists.size == 0:
                # No valid distances, return np.nan
                eta_row = np.nan
            else:
                eta_row = np.quantile(quantile_row_dists, eta_row)
            if quantile_col_dists.size == 0:
                # No valid distances, return np.nan
                eta_col = np.nan
            else:
                eta_col = np.quantile(quantile_col_dists, eta_col)

        # Find the row nearest neighbors indexes
        row_nearest_neighbors = np.nonzero(row_dists <= eta_row)[0]

        # Find the col nearest neighbors indexes
        col_nearest_neighbors = np.nonzero(col_dists <= eta_col)[0]

        # neighbors can only be used if they are observed
        row_nearest_neighbors = row_nearest_neighbors[
            mask_array[row_nearest_neighbors, column] == 1
        ]
        col_nearest_neighbors = col_nearest_neighbors[
            mask_array[row, col_nearest_neighbors] == 1
        ]

        # Use doubly robust nearest neighbors to combine row and col
        y_itprime = data_array[row, col_nearest_neighbors]
        y_jt = data_array[row_nearest_neighbors, column]

        if len(y_itprime) == 0 and len(y_jt) == 0:
            return np.array(np.nan)

        # get intersecting entries
        j_inds, tprime_inds = np.meshgrid(
            row_nearest_neighbors, col_nearest_neighbors, indexing="ij"
        )

        y_jtprime = data_array[j_inds, tprime_inds]
        mask_jtprime = mask_array[j_inds, tprime_inds]

        # nonzero gets all indices of the mask that are 1
        intersec_inds = np.nonzero(mask_jtprime == 1)

        y_itprime_inter = y_itprime[
            intersec_inds[1]
        ]  # this is a array of column values for all intersecting triplets
        y_jt_inter = y_jt[
            intersec_inds[0]
        ]  # this is a array of row values for all intersecting triplets
        # note: defaults to rownn if no intersection -> should default to ts-nn instead?
        if len(y_itprime_inter) == 0 or len(y_jt_inter) == 0:
            return np.array(
                data_type.average(y_jt)
                if len(y_jt) > 0
                else data_type.average(y_itprime)
            )
        sum_y = y_itprime_inter + y_jt_inter - y_jtprime[intersec_inds]
        avg = data_type.average(sum_y)
        return avg

    def _calculate_distances(
        self,
        row: int,
        col: int,
        data_array: npt.NDArray,
        mask_array: npt.NDArray,
        data_type: DataType,
    ) -> None:
        """Sets the distances for the imputer.
        Sets the distances as a class attribute, so returns nothing.

        Args:
            row (int): Row index
            col (int): Column index
            data_array (npt.NDArray): Data matrix
            mask_array (npt.NDArray): Mask matrix
            data_type (DataType): Data type to use (e.g. scalars, distributions)

        """
        data_shape = data_array.shape
        n_rows = data_shape[0]
        n_cols = data_shape[1]

        if row not in self.row_distances:
            # Calculate distances between rows
            row_dists = np.zeros((n_rows, n_cols))

            # Scalar optimization with vectorized operations instead of loops
            if isinstance(data_type, Scalar):
                # Determine overlap columns for any pairwise rows
                overlap_columns_mask = np.logical_and(
                    np.tile(mask_array[row], (n_rows, 1)), mask_array
                )
                row_big_matrix = np.tile(data_array[row], (n_rows, 1))
                row_dists = np.power(data_array - row_big_matrix, 2).astype(np.float64)
                # We need the row dists as a float matrix to use np.nanmean
                row_dists[~overlap_columns_mask] = np.nan
            else:
                for i in range(n_rows):
                    # Get columns observed in both row i and row
                    overlap_columns = np.logical_and(mask_array[row], mask_array[i])

                    if not np.any(overlap_columns):
                        row_dists[i, :] = np.nan
                        continue

                    # Calculate distance between rows
                    for j in range(n_cols):
                        if not overlap_columns[
                            j
                        ]:  # Skip missing values and the target column
                            row_dists[i, j] = np.nan
                        else:
                            row_dists[i, j] = data_type.distance(
                                data_array[row, j], data_array[i, j]
                            )
            self.row_distances[row] = row_dists
            # self.row_distances[row][row] = np.inf  # Exclude the row itself

        if col not in self.col_distances:
            # Calculate distances between columns
            col_dists = np.zeros((n_rows, n_cols))

            # Scalar optimization with vectorized operations instead of loops
            if isinstance(data_type, Scalar):
                # Determine overlap columns for any pairwise rows
                overlap_columns_mask = np.logical_and(
                    np.tile(mask_array[:, col].reshape(-1, 1), (1, n_cols)), mask_array
                )
                row_big_matrix = np.tile(data_array[:, col].reshape(-1, 1), (1, n_cols))
                col_dists = np.power(data_array - row_big_matrix, 2).astype(np.float64)
                # We need the col dists as a float matrix to use np.nanmean
                col_dists[~overlap_columns_mask] = np.nan

            else:
                for j in range(n_cols):
                    # Get rows observed in both row i and row
                    overlap_columns = np.logical_and(
                        mask_array[:, col], mask_array[:, j]
                    )

                    if not np.any(overlap_columns):
                        col_dists[:, j] = np.nan
                        continue

                    # Calculate distance between columns
                    for i in range(n_rows):
                        if not overlap_columns[
                            i
                        ]:  # Skip missing values and the target column
                            col_dists[i, j] = np.nan
                        else:
                            col_dists[i, j] = data_type.distance(
                                data_array[i, col], data_array[i, j]
                            )
            self.col_distances[col] = col_dists
            # self.col_distances[col][col] = np.inf


class TSEstimator(EstimationMethod):
    """Estimate the missing value using two-sided nearest neighbors.

    This method first finds the row and column neighborhoods (based on a
    distance computed over overlapping observed entries, excluding the target)
    and then imputes the missing entry by averaging the observed values
    over the cross-product of these neighborhoods.
    """

    def __init__(self, is_percentile: bool = True):
        """Initialize the two-sided estimator.

        Args:
            is_percentile (bool): Whether to use percentile-based threshold. Defaults to True.

        """
        super().__init__(is_percentile)
        self.estimator = DREstimator(is_percentile=is_percentile)

    def __str__(self):
        return "TSEstimator"

    def impute(
        self,
        row: int,
        column: int,
        data_array: npt.NDArray,
        mask_array: npt.NDArray,
        distance_threshold: Union[float, Tuple[float, float]],
        data_type: DataType,
        allow_self_neighbor: bool = True,
        **kwargs: Any,
    ) -> npt.NDArray:
        r"""Impute the missing value at the given row and column using two-sided NN.

        Args:
            row (int): Target row index.
            column (int): Target column index.
            data_array (npt.NDArray): Data matrix.
            mask_array (npt.NDArray): Boolean mask matrix (True if observed).
            distance_threshold (float or Tuple[float, float]): Distance threshold(s) for row and column neighborhoods. This is our \vec eta = (\eta_row, \eta_col).
            data_type (DataType): Provides methods for computing distances and averaging.
            allow_self_neighbor (bool): Whether to allow self-neighbor. Defaults to False.
            **kwargs (Any): Additional keyword arguments

        Returns:
            npt.NDArray: Imputed value.

        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            self._calculate_distances(row, column, data_array, mask_array, data_type)
            all_dists = np.copy(self.estimator.row_distances[row])
            # Exclude the target column
            if not allow_self_neighbor:
                all_dists[:, column] = np.nan
            row_dists = np.nanmean(all_dists, axis=1)
            if not mask_array[row, column] and not allow_self_neighbor:
                row_dists[row] = np.inf  # Exclude the target row
            # Exclude the target row
            all_dists = np.copy(self.estimator.col_distances[column])
            if not allow_self_neighbor:
                all_dists[row, :] = np.nan
            col_dists = np.nanmean(all_dists, axis=0)
            if not mask_array[row, column] and not allow_self_neighbor:
                col_dists[column] = np.inf  # Exclude the target col

        if isinstance(distance_threshold, tuple):
            eta_row, eta_col = distance_threshold
        else:
            eta_row = distance_threshold
            eta_col = distance_threshold
        if self.is_percentile:
            # NOTE: we assume eta_row and eta_col are in [0, 1] in this case
            quantile_row_dists = row_dists[~np.isnan(row_dists) & (row_dists != np.inf)]
            quantile_col_dists = col_dists[~np.isnan(col_dists) & (col_dists != np.inf)]
            if quantile_row_dists.size == 0:
                # No valid distances, return np.nan
                eta_row = np.nan
            else:
                eta_row = np.quantile(quantile_row_dists, eta_row)
            if quantile_col_dists.size == 0:
                # No valid distances, return np.nan
                eta_col = np.nan
            else:
                eta_col = np.quantile(quantile_col_dists, eta_col)

        # Establish the neighborhoods subject to the distance thresholds
        row_nearest_neighbors = np.where(row_dists <= eta_row)[0]
        # This is the set N_row(i, j) = {i' | d^2(i, i') <= eta_row^2}
        col_nearest_neighbors = np.where(col_dists <= eta_col)[0]
        # This is the set N_col(i, j) = {j' | d^2(j, j') <= eta_col^2}
        neighborhood_submatrix = data_array[
            np.ix_(row_nearest_neighbors, col_nearest_neighbors)
        ]
        # This is the submatrix of the cross-product of the row and column neighborhoods
        mask_array = mask_array.astype(bool)  # for efficient indexing
        neighborhood_mask = mask_array[
            np.ix_(row_nearest_neighbors, col_nearest_neighbors)
        ]
        # This is the mask of the cross-product of the row and column neighborhoods

        values_for_estimation = neighborhood_submatrix[neighborhood_mask.astype(bool)]
        if values_for_estimation.size == 0:
            if mask_array[row, column]:
                logger.log(
                    logging.WARNING,
                    f"Warning: No valid neighbors found for ({row}, {column}). Returning observed value.",
                )
                return data_array[row, column]
            else:
                logger.log(
                    logging.WARNING,
                    f"Warning: No valid neighbors found for ({row}, {column}). Returning np.nan.",
                )
                return np.array(np.nan)
        else:
            theta_hat = data_type.average(values_for_estimation)
            return theta_hat

    def _calculate_distances(
        self,
        row: int,
        col: int,
        data_array: npt.NDArray,
        mask_array: npt.NDArray,
        data_type: DataType,
    ) -> None:
        """Sets the distances for the imputer.
        Sets the distances as a class attribute, so returns nothing.

        Args:
            row (int): Row index
            col (int): Column index
            data_array (npt.NDArray): Data matrix
            mask_array (npt.NDArray): Mask matrix
            data_type (DataType): Data type to use (e.g. scalars, distributions)

        """
        self.estimator._calculate_distances(row, col, data_array, mask_array, data_type)
        # We use the DREstimator class to calculate the distances
        # because it is the same as the two-sided estimator


class AWNNEstimator(EstimationMethod):
    """Estimate the missing value using AWNN."""

    def __init__(
        self,
        delta: float = 1,
        noise_variance: Optional[
            float
        ] = None,  # this is a the new variable specific to this!
        convergence_threshold: float = 1e-4,
        max_iterations: int = 10,
    ):
        """Initialize the AWNN estimator."""
        # sanity checks:
        if delta < 0 or delta > 1:
            raise ValueError("Delta must be between 0 and 1.")
        if convergence_threshold < 0:
            raise ValueError("Convergence threshold must be non-negative.")
        if max_iterations <= 0:
            raise ValueError("Max iterations must be positive.")
        if noise_variance is not None and noise_variance < 0:
            raise ValueError("Noise variance must be non-negative.")
        self.row_distances = np.array([])
        self.noise_variance = noise_variance
        self.convergence_threshold = convergence_threshold
        self.max_iterations = max_iterations
        self.delta = delta
        self.estimated_signal_matrix = None  # these 2 are used to cache full matrix result since impute is 1 (r,c) at a time
        self.delta_value_for_signal_matrix = None

    def __str__(self):
        return "AWNNEstimator"

    def _impute_single_value_helper(
        self,
        row: int,
        column: int,
        data_array: npt.NDArray,
        mask_array: npt.NDArray,
        data_type: DataType,
        vectorize: bool = True,
    ) -> npt.NDArray:
        """Imputes one specific value using the AWNN method."""
        n_rows, n_cols = data_array.shape
        delta = self.delta / np.sqrt(n_rows)
        logger.info("delta: %s" % delta)  # TODO switch to logger.log
        logger.info(
            "noise_variance: %s" % self.noise_variance
        )  # TODO switch to logger.log
        if self.noise_variance is None:
            noise_variance = np.var(data_array[mask_array == 1]) / 2
            self.noise_variance = noise_variance
        else:
            noise_variance = self.noise_variance

        observed_rows = np.where(mask_array[:, column] == 1)[0]
        n_observed = len(observed_rows)
        if n_observed == 0:
            return np.array(np.nan)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            # Calculate distances between rows
            if not self.row_distances.size:
                self._calculate_distances(
                    row, column, data_array, mask_array, data_type
                )
            row_distances = self.row_distances

        d = row_distances[row, observed_rows]
        d = np.where(observed_rows == row, 0, d - 2 * noise_variance)
        row_dist_min = min(0, np.min(d))
        d = np.where(observed_rows == row, 0, d - row_dist_min)

        mean_distance = np.mean(d)
        dist_diff = d - mean_distance
        # print (noise_variance)
        if noise_variance != 0:
            weights = (1 / n_observed) - dist_diff / (
                8 * noise_variance * np.log(1 / delta)
            )
        else:
            weights = (1 / n_observed) - dist_diff / (8 * np.log(1 / delta))
        sorted_weights = np.sort(weights)[::-1]
        if vectorize:  # datatype-independent optimization!
            cumsum_weights = np.cumsum(sorted_weights)
            threshold = (cumsum_weights - 1) / np.arange(1, n_observed + 1)
            masker = sorted_weights > threshold
            u = threshold[masker][-1] if masker.any() else 0
            weights = np.maximum(0, weights - u)
        else:
            weight_sum = 0
            u = 0
            for k in range(n_observed):
                weight_sum += sorted_weights[k]
                u_new = (weight_sum - 1) / (k + 1)
                if k == n_observed - 1 or sorted_weights[k + 1] <= u_new:
                    u = u_new
                    break
            weights = np.maximum(0, weights - u)
            # print("weights for row %d, column %d: %s" % (row, column, weights))
        ret_val = np.sum(weights * data_array[observed_rows, column])
        return ret_val

    def _fit_full_matrix(
        self,
        data_array: npt.NDArray,
        mask_array: npt.NDArray,
        data_type: DataType,
    ) -> np.ndarray:
        n_rows, n_cols = data_array.shape
        imputed_data = np.zeros_like(data_array)
        for iter in range(self.max_iterations):
            logger.info("Iteration %d" % iter)  # TODO switch to logger.log
            for i in range(n_rows):
                for j in range(n_cols):
                    imputed_data[i, j] = self._impute_single_value_helper(
                        i, j, data_array, mask_array, data_type
                    )
            diff = imputed_data[mask_array == 1] - data_array[mask_array == 1]
            diff = diff[~np.isnan(diff)]  # Remove any NaN values
            if len(diff) > 0:
                new_variance_estimate = np.var(diff)
                if self.noise_variance is None:
                    raise ValueError(
                        "Noise variance is not set, which should not happen."
                    )
                if (
                    abs(new_variance_estimate - self.noise_variance)
                    / self.noise_variance
                    < self.convergence_threshold
                ):
                    logger.info(
                        f"Converged after {iter + 1} iterations and final noise variance: {new_variance_estimate}"
                    )
                    self.noise_variance = new_variance_estimate
                    break
                self.noise_variance = new_variance_estimate
        self.estimated_signal_matrix = imputed_data
        return imputed_data

    def impute(
        self,
        row: int,
        column: int,
        data_array: npt.NDArray,
        mask_array: npt.NDArray,
        distance_threshold: Union[float, Tuple[float, float]],  # unused
        data_type: DataType,
        allow_self_neighbor: bool = False,
        **kwargs: Any,
    ) -> npt.NDArray:
        """Impute the missing value at the given row and column.

        Args:
            row (int): Row index of the missing value.
            column (int): Column index of the missing value.
            data_array (npt.NDArray): Data matrix containing observed and missing values.
            mask_array (npt.NDArray): Boolean mask matrix indicating observed values.
            distance_threshold (Union[float, Tuple[float, float]]): Distance threshold (unused in this method).
            allow_self_neighbor (bool): Whether to allow self-neighbor. Defaults to False. (unused in this method)
            data_type (DataType): Data type providing methods for distance calculation and averaging.
            allow_self_neighbor (bool): Whether to allow self-neighbor. Defaults to False. (unused in this method)
            **kwargs (Any): Additional keyword arguments.

        Returns:
            npt.NDArray: Imputed value for the specified row and column.

        """
        full_matrix = self.impute_all(
            data_array,
            mask_array,
            distance_threshold,
            data_type,
        )
        val_at_index = full_matrix[row, column]
        return val_at_index

    def impute_all(
        self,
        data_array: npt.NDArray,
        mask_array: npt.NDArray,
        distance_threshold: Union[float, Tuple[float, float]],
        data_type: DataType,
    ) -> npt.NDArray:
        """Impute all missing values in the data array.

        Args:
            data_array (npt.NDArray): Data matrix containing observed and missing values.
            mask_array (npt.NDArray): Boolean mask matrix indicating observed values.
            distance_threshold (Union[float, Tuple[float, float]]): Distance threshold for nearest neighbors.
            data_type (DataType): Data type providing methods for distance calculation and averaging.

        Returns:
            npt.NDArray: Imputed value for the specified row and column.

        """
        if self.estimated_signal_matrix is None:
            full_mat = self._fit_full_matrix(
                data_array,
                mask_array,
                data_type,
            )
            self.estimated_signal_matrix = full_mat  # cache it
        else:
            full_mat = self.estimated_signal_matrix

        return full_mat

    def _calculate_distances(
        self,
        row: int,
        col: int,
        data_array: npt.NDArray,
        mask_array: npt.NDArray,
        data_type: DataType,
        vectorize: bool = True,
    ) -> None:
        """Computes distances, caches them. Vectorized version, tested for equivalence."""
        if vectorize and isinstance(data_type, Scalar):
            X = data_array
            M = mask_array.astype(float)
            X_masked = X * M

            # count of overlapping columns for each pair
            counts = M @ M.T

            # A[i,j] = Σₖ Mᵢₖ·Mⱼₖ·Xᵢₖ²
            A = (X_masked**2) @ M.T
            # C[i,j] = Σₖ Mᵢₖ·Mⱼₖ·Xᵢₖ·Xⱼₖ
            C = X_masked @ X_masked.T

            D = A + A.T - 2 * C
            # normalize, inf where no overlap
            with np.errstate(divide="ignore", invalid="ignore"):
                row_distances = D / counts
            row_distances[counts == 0] = np.inf
            np.fill_diagonal(row_distances, 0)
        else:
            n_rows, n_cols = data_array.shape
            row_distances = np.zeros((n_rows, n_rows))

            for i in tqdm(range(n_rows)):
                for j in range(i + 1, n_rows):
                    overlap_cols = np.logical_and(mask_array[i, :], mask_array[j, :])
                    if not np.any(overlap_cols):
                        row_distances[i, j] = np.inf
                        row_distances[j, i] = np.inf
                        continue
                    if isinstance(data_type, Scalar):
                        row_distances[i, j] = np.sum(
                            np.square(
                                data_array[i, overlap_cols]
                                - data_array[j, overlap_cols]
                            )
                        )
                    else:
                        for k in range(n_cols):
                            if not overlap_cols[k]:
                                continue
                            row_distances[i, j] += data_type.distance(
                                data_array[i, k], data_array[j, k]
                            )
                    row_distances[i, j] /= np.sum(overlap_cols)
                    row_distances[j, i] = row_distances[i, j]

        self.row_distances = row_distances


class AutoEstimator(EstimationMethod):
    """Estimate the missing entry using "Auto-NN" idea (Kyuseong Choi)."""

    def __init__(self, is_percentile: bool = True):
        super().__init__(is_percentile)
        self.row_distances = dict()
        self.col_distances = dict()
        # insert default f and g functions
        # self.f = lambda x: x
        # self.g = lambda x: x
        self.alpha = 1.0
        self.drnn_imputer = DREstimator(is_percentile=is_percentile)
        self.ts_imputer = TSEstimator(is_percentile=is_percentile)

    def impute(
        self,
        row: int,
        column: int,
        data_array: npt.NDArray,
        mask_array: npt.NDArray,
        distance_threshold: Union[float, Tuple[float, float]],
        data_type: DataType,
        allow_self_neighbor: bool = False,
        **kwargs: Any,
    ) -> npt.NDArray:
        """Impute the missing value at the given row and column using doubly robust method.

        Args:
        ----
            row (int): Row index
            column (int): Column index
            data_array (npt.NDArray): Data matrix
            mask_array (npt.NDArray): Mask matrix
            distance_threshold (float or Tuple[float, float]): Distance threshold for nearest neighbors
            or a tuple of (row_threshold, col_threshold) for row and column respectively.
            data_type (DataType): Data type to use (e.g. scalars, distributions)
            allow_self_neighbor (bool): Whether to allow self-neighbor. Defaults to False.
            **kwargs (Any): Additional keyword arguments

        """
        if self.alpha is None and "alpha" not in kwargs:
            raise TypeError(
                "AutoEstimator.impute() missing 1 required keyword-only argument: 'alpha'"
            )
        alpha = self.alpha if self.alpha else kwargs.pop("alpha")
        # TODO: handle allow self neighbor (Caleb)
        drnn_impute = self.drnn_imputer.impute(
            row,
            column,
            data_array,
            mask_array,
            distance_threshold,
            data_type,
            allow_self_neighbor=False,
        )
        ts_impute = self.ts_imputer.impute(
            row,
            column,
            data_array,
            mask_array,
            distance_threshold,
            data_type,
            allow_self_neighbor=True,
        )

        # interpolate between tsnn and drnn
        final_impute = (1 - alpha) * drnn_impute + alpha * ts_impute

        return final_impute

    def _calculate_distances(
        self,
        row: int,
        col: int,
        data_array: npt.NDArray,
        mask_array: npt.NDArray,
        data_type: DataType,
    ) -> None:
        """Sets the distances for the imputer.
        Sets the distances as a class attribute, so returns nothing.

        Args:
            row (int): Row index
            col (int): Column index
            data_array (npt.NDArray): Data matrix
            mask_array (npt.NDArray): Mask matrix
            data_type (DataType): Data type to use (e.g. scalars, distributions)

        """
        data_shape = data_array.shape
        n_rows = data_shape[0]
        n_cols = data_shape[1]

        if row not in self.row_distances:
            # Calculate distances between rows
            row_dists = np.zeros((n_rows, n_cols))

            for i in range(n_rows):
                # Get columns observed in both row i and row
                overlap_columns = np.logical_and(mask_array[row], mask_array[i])

                if not np.any(overlap_columns):
                    row_dists[i, :] = np.nan
                    continue

                # Calculate distance between rows
                for j in range(n_cols):
                    if not overlap_columns[
                        j
                    ]:  # Skip missing values and the target column
                        row_dists[i, j] = np.nan
                    else:
                        row_dists[i, j] = data_type.distance(
                            data_array[row, j], data_array[i, j]
                        )
            self.row_distances[row] = row_dists

        if col not in self.col_distances:
            # Calculate distances between columns
            col_dists = np.zeros((n_rows, n_cols))

            for j in range(n_cols):
                # Get rows observed in both row i and row
                overlap_columns = np.logical_and(mask_array[:, col], mask_array[:, j])

                if not np.any(overlap_columns):
                    col_dists[:, j] = np.nan
                    continue

                # Calculate distance between columns
                for i in range(n_rows):
                    if not overlap_columns[
                        i
                    ]:  # Skip missing values and the target column
                        col_dists[i, j] = np.nan
                    else:
                        col_dists[i, j] = data_type.distance(
                            data_array[i, col], data_array[i, j]
                        )
            self.col_distances[col] = col_dists



"""Confounded Staggered adoption simulation

T: Number of cols
n: Number of measurements per entry
d: Dimension of measurements
beta: Degree of confounding

"""

import numpy as np
import math as math
from typing import Tuple


def expit(x: float) -> float:
    """Compute the logistic sigmoid function of x.

    Args:
        x: Input value

    Returns:
        float: 1 / (1 + exp(-x))

    """
    return 1 / (1 + np.exp(-x))


def gendata_s_adopt(
    N: int, T: int, n: int, d: int, beta: Tuple[float, float], seed: int
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Generates confounded/staggered adoption "Gaussian" data for distributional NN, with latent dimension r = 2

    Args:
        N: Number of rows (even)
        T: Number of columns
        n: Number of measurements per entry
        d: Dimension of measurements (even)
        beta: vector of two fractions between (0, 1)
        seed: random seed

    Returns:
        Data: (N, T, n, d) array of data
        Masking: (N, T) array of missing indicators
        true_Mean: (N, T, d) array of true means
        true_Cov: (N, T, d, d) array of true covariances

    """
    np.random.seed(seed=seed)

    Data = np.zeros((N, T))
    true_Mean = np.zeros((N, T))
    true_Cov = np.zeros((N, T))

    u_1 = np.random.uniform(-1, 1, N)
    u_2 = np.random.uniform(0.2, 1, N)

    v_1 = np.random.uniform(-2, 2, T)
    v_2 = np.random.uniform(0.5, 2, T)

    even_ones = np.repeat([0, 1], int(d / 2))
    odd_ones = np.repeat([1, 0], int(d / 2))

    for i in range(N):
        for t in range(T):
            m_it = u_1[i] * v_1[t] * (even_ones - odd_ones)
            c_it = np.diag(u_2[i] * v_2[t] * (0.5 * even_ones + odd_ones))
            true_Mean[i, t] = m_it
            true_Cov[i, t] = c_it
            dat_mat = np.random.multivariate_normal(m_it, c_it, size=n)
            Data[i, t] = dat_mat

    Masking = np.zeros((N, T))
    pre_Masking = np.zeros((N, T))

    g1_inds = np.arange(0, N // 3)
    g2_inds = np.arange(N // 3, 2 * N // 3)
    g3_inds = np.arange(2 * N // 3, N)

    gamma_1 = [2, 0.7, 1, 0.7]
    gamma_2 = [2, 0.2, 1, 0.2]

    T1_lower = math.floor(T ** beta[0])
    T2_lower = math.floor(T ** beta[1])

    for i in range(N):
        if i in g1_inds:
            pre_Masking[i, :] = np.concatenate(
                (np.ones(T1_lower), np.zeros(T - T1_lower))
            )
            for t in range(T - T1_lower):
                pre_Masking[i, (t + T1_lower)] = np.random.binomial(
                    1,
                    expit(
                        gamma_1[0]
                        + (0.99**t) * gamma_1[1] * u_1[i - 1]
                        + gamma_1[2] * u_1[i]
                        + (0.99**t) * gamma_1[3] * u_1[i + 1]
                    ),
                    1,
                )
            pre_A = pre_Masking[i, :]
            if len([i for i in range(len(pre_A)) if pre_A[i] == 0]) == 0:
                Masking[i, :] = pre_A
            elif len([i for i in range(len(pre_A)) if pre_A[i] == 0]) > 0:
                adopt_time = min([i for i in range(len(pre_A)) if pre_A[i] == 0])
                Masking[i, :] = np.concatenate(
                    (np.ones(adopt_time), np.zeros(T - adopt_time))
                )
        elif i in g2_inds:
            pre_Masking[i, :] = np.concatenate(
                (np.ones(T2_lower), np.zeros(T - T2_lower))
            )
            for t in range(T - T2_lower):
                pre_Masking[i, (t + T2_lower)] = np.random.binomial(
                    1,
                    expit(
                        gamma_2[0]
                        + (1.01**t) * gamma_2[1] * u_1[i - 1]
                        + gamma_2[2] * u_1[i]
                        + (1.01**t) * gamma_2[3] * u_1[i + 1]
                    ),
                    1,
                )
            pre_A = pre_Masking[i, :]
            if len([i for i in range(len(pre_A)) if pre_A[i] == 0]) == 0:
                Masking[i, :] = pre_A
            elif len([i for i in range(len(pre_A)) if pre_A[i] == 0]) > 0:
                adopt_time = min([i for i in range(len(pre_A)) if pre_A[i] == 0])
                Masking[i, :] = np.concatenate(
                    (np.ones(adopt_time), np.zeros(T - adopt_time))
                )
        elif i in g3_inds:
            Masking[i, :] = np.ones(T)

    return Data, Masking, true_Mean, true_Cov



"""Dataset loader for the Tax Burden on Tobacco dataset.

Source: Centers for Disease Control and Prevention (CDC)
License: Public Domain
Paper Reference for Prop99:
    Abadie, A., Diamond, A., & Hainmueller, J. (2010).
    Synthetic Control Methods for Comparative Case Studies: Estimating the Effect of California's Tobacco Control Program.
    Journal of the American Statistical Association, 105(490), 493-505.
    https://doi.org/10.1198/jasa.2009.ap08746
"""

from nsquared.datasets.dataloader_base import NNDataLoader
from nsquared.datasets.dataloader_factory import register_dataset
from nsquared.data_types import DataType
import numpy as np
import pandas as pd
from typing import Any
import logging
from joblib import Memory
import os
import requests

memory = Memory(".joblib_cache", verbose=2)

logger = logging.getLogger(__name__)

params = {
    "start_year": (
        int,
        1970,
        "Start year for the data range.",
    ),
    "end_year": (
        int,
        2019,
        "End year for the data range.",
    ),
    "sample_states": (
        int,
        None,
        "Number of states to sample from the dataset. By default, returns all.",
    ),
    "seed": (int, None, "Random seed for reproducibility"),
}


@register_dataset("prop99", params)
class Prop99DataLoader(NNDataLoader):
    """Data from the Tax Burden on Tobacco dataset, focusing on cigarette consumption in packs.
    To initialize with default settings, use: NNData.create("prop99").
    """

    urls = {
        "tobacco_data": "https://data.cdc.gov/api/views/7nwe-3aj9/rows.csv?accessType=DOWNLOAD",
    }

    # 12 states that implemented Prop99 or similar legislation after 1988
    TREATED_STATES = {
        "CA",  # California
        "MA",  # Massachusetts
        "AZ",  # Arizona
        "OR",  # Oregon
        "FL",  # Florida
        "AK",  # Alaska
        "HI",  # Hawaii
        "MD",  # Maryland
        "MI",  # Michigan
        "NJ",  # New Jersey
        "NY",  # New York
        "WA",  # Washington
    }
    # 38 states that never implemented Prop99 or similar legislation
    CONTROL_STATES = {
        "TX",
        "WI",
        "MT",
        "RI",
        "KS",
        "ME",
        "UT",
        "VA",
        "IN",
        "GA",
        "AL",
        "SD",
        "NH",
        "DE",
        "NC",
        "CO",
        "AR",
        "CT",
        "MN",
        "NV",
        "LA",
        "IA",
        "NE",
        "SC",
        "OH",
        "TN",
        "WV",
        "KY",
        "ID",
        "MS",
        "IL",
        "WY",
        "VT",
        "ND",
        "PA",
        "OK",
        "MO",
        "NM",
    }
    # Rows to exclude from the dataset entirely
    EXCLUDED_STATES = {"DC"}

    def __init__(
        self,
        start_year: int = 1970,
        end_year: int = 2000,
        state: str = "CA",
        seed: int | None = None,
        **kwargs: Any,
    ):
        """Initializes the Prop99 data loader.

        Args:
        ----
            start_year: Start year for the data range. Default: 1970.
            end_year: End year for the data range. Default: 2019.
            state: State to impute. Default: "CA".
            seed: Random seed for reproducibility. Default: None
            kwargs: Additional keyword arguments.

        """
        super().__init__(
            **kwargs,
        )
        self.start_year = start_year
        self.end_year = end_year
        self.state = state
        self.data = None
        self.mask = None
        if seed is not None:
            np.random.seed(
                seed
            )  # instantiate random seed if provided but do it only once here

    def process_data_scalar(self) -> tuple[np.ndarray, np.ndarray]:
        """Process the data into scalar setting focusing on cigarette consumption in packs.

        Returns
        -------
            data: 2d processed data matrix of floats (states × years)
            mask: Mask for processed data

        """
        df = self._load_data()

        # Filter for cigarette consumption in packs
        logger.info(f"Original data shape: {df.shape}")
        df_consumption = df[
            df["SubMeasureDesc"] == "Cigarette Consumption (Pack Sales Per Capita)"
        ]
        logger.info(f"After SubMeasureDesc filter: {df_consumption.shape}")

        # Filter by year range
        df_consumption = df_consumption[
            (df_consumption["Year"] >= self.start_year)
            & (df_consumption["Year"] <= self.end_year)
        ]
        logger.info(f"After year range filter: {df_consumption.shape}")

        # Make sure we're working with a pandas DataFrame
        df_consumption = (
            pd.DataFrame(df_consumption)
            if not isinstance(df_consumption, pd.DataFrame)
            else df_consumption
        )

        # Pivot the data to get states as rows and years as columns
        data_df = df_consumption.pivot(
            index="LocationAbbr", columns="Year", values="Data_Value"
        )
        logger.info(f"Pivoted data shape: {data_df.shape}")

        # Retain only the control states and the treated state
        rows_to_drop = list(
            self.TREATED_STATES.union(self.EXCLUDED_STATES) - {self.state}
        )
        data_df = data_df.drop(rows_to_drop)
        logger.info(f"After excluding states: {data_df.shape}")

        # Create the mask: 0 for CA after 1988, 1 for everything else
        mask_df = pd.DataFrame(1, index=data_df.index, columns=data_df.columns)

        # If the treated state is in the data, set mask values to 0 for years after 1988
        assert self.state in mask_df.index, (
            f"{self.state} is not a valid state in the data"
        )
        # reorder the rows so that the treated state is the first row
        data_df = data_df.reindex(
            [self.state] + [state for state in data_df.index if state != self.state]
        )
        mask_df = mask_df.reindex(
            [self.state] + [state for state in mask_df.index if state != self.state]
        )

        ca_years = [year for year in mask_df.columns if int(year) > 1988]
        mask_df.loc[self.state, ca_years] = 0

        # Convert to numpy arrays
        data = data_df.to_numpy()
        mask = mask_df.to_numpy()

        # Store the processed data
        self.data = data
        self.mask = mask
        self.state_names = data_df.index.tolist()
        self.years = data_df.columns.tolist()

        return data, mask

    def process_data_distribution(
        self, data_type: DataType | None = None
    ) -> tuple[np.ndarray, np.ndarray]:
        """Does not apply to Prop99 as it is not a distributional dataset.

        Args:
            data_type: Data type to process. Default: None.

        """
        raise ValueError("There is no distributional data for Prop99.")

    def get_full_state_as_dict(self, include_metadata: bool = False) -> dict:
        """Returns the full state as a dictionary. For Prop99, this includes the data, masking matrix, and the custom parameters.

        If the data and mask are None, then the data has not been processed yet. Call process_data_scalar() to process the data first.

        Args:
            include_metadata (bool): Whether to include metadata in the dictionary. Default: False.

        """
        full_state = {
            "data": self.data,
            "mask": self.mask,
            "custom_params": {
                "start_year": self.start_year,
                "end_year": self.end_year,
            },
        }

        if include_metadata and hasattr(self, "state_names") and hasattr(self, "years"):
            full_state["metadata"] = {
                "state_names": self.state_names,
                "years": self.years,
                "data_description": "Cigarette Consumption (Per Capita in packs)",
            }

        return full_state

    @classmethod
    @memory.cache
    def _load_data(cls) -> pd.DataFrame:
        """Download and load the Tax Burden on Tobacco dataset."""
        csv_path = "tobacco_data.csv"

        # Check for locally provided file
        local_file = "The_Tax_Burden_on_Tobacco.csv"
        if os.path.exists(local_file):
            logger.info(f"Using locally provided file: {local_file}")
            return pd.read_csv(local_file)

        # Download if missing
        needs_download = not os.path.exists(csv_path)

        if needs_download:
            logger.info("Downloading Tax Burden on Tobacco dataset...")
            response = requests.get(cls.urls["tobacco_data"])
            if response.status_code != 200:
                raise RuntimeError(
                    f"Failed to download dataset: {response.status_code}"
                )
            # Write the content directly instead of using raw stream
            with open(csv_path, "wb") as f:
                f.write(response.content)

        logger.info(f"Loading data from {csv_path}")
        df = pd.read_csv(csv_path)
        return df



from .nnimputer import DataType
import numpy.typing as npt
from typing import Any, Callable
import numpy as np
import warnings


class Scalar(DataType):
    """Data type for scalars."""

    def distance(self, obj1: float, obj2: float) -> float:
        """Calculate the distance between two scalars.

        Args:
            obj1 (float): Scalar 1
            obj2 (float): Scalar 2

        Returns:
            float: Distance between the two scalars

        """
        return (obj1 - obj2) ** 2

    def average(self, object_list: npt.NDArray[Any]) -> Any:
        """Calculate the average of a list of scalars.

        Args:
            object_list (list[Any]): List of scalars

        Returns:
            Any: Average of the scalars

        """
        # Compute the mean using only the non-nan values.
        # If all values are nan, return nan.
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            return np.nanmean(object_list)


class DistributionKernelMMD(DataType):
    """Data type for distributions using Kernel MMD."""

    def __init__(self, kernel: str, tuning_parameter: float = 0.5, d: int = 1):
        """Initialize the distribution data type with a kernel.

        Args:
            kernel (str): Kernel to use for the MMD
            tuning_parameter (float): Inverse bandwidth parameter for the exponential kernel
            d (int): Dimension of the data

        """
        supported_kernels = ["linear", "square", "exponential"]

        if kernel not in supported_kernels:
            raise ValueError(
                f"Kernel {kernel} is not supported. Supported kernels are {supported_kernels}"
            )

        self.kernel = kernel
        self.tuning_parameter = tuning_parameter
        self.d = d

    def distance(self, obj1: npt.NDArray, obj2: npt.NDArray) -> float:
        """Calculate the distance between two distributions using Kernel MMD.

        Args:
            obj1 (npt.NDArray): (n, d) array of n samples of dimension d
            obj2 (npt.NDArray): (m, d) array of m samples of dimension d

        Returns:
            float: U-statistic of the squared MMD distance between the two distributions

        """
        m = obj1.shape[0]
        n = obj2.shape[0]

        if len(obj1.shape) == 1:
            obj1 = obj1[:, np.newaxis]
        if len(obj2.shape) == 1:
            obj2 = obj2[:, np.newaxis]

        assert obj1.shape[1] == obj2.shape[1]

        XX = np.matmul(obj1, np.transpose(obj1))  # m by m matrix with x_i^Tx_j
        YY = np.matmul(obj2, np.transpose(obj2))  # n by n matrix with y_i^Ty_j
        XY = np.matmul(obj1, np.transpose(obj2))  # m by n matrix with x_i^Ty_j

        if self.kernel == "linear":
            kXX, kYY, kXY = XX, YY, XY
        elif self.kernel == "square":
            kXX, kYY, kXY = (
                (XX + np.ones((m, m))) ** 2,
                (YY + np.ones((n, n))) ** 2,
                (XY + np.ones((m, n))) ** 2,
            )
        elif self.kernel == "exponential":
            dXX_mm = np.vstack(
                (np.diag(XX),) * m
            )  # m*m matrix : each row is the diagonal x_i^Tx_i
            dYY_nn = np.vstack(
                (np.diag(YY),) * n
            )  # n*n matrix : each row is the diagonal y_i^Ty_i
            dXX_mn = np.vstack(
                (np.diag(XX),) * n
            ).transpose()  # m*n matrix : each row is the diagonal x_i^Tx_i
            dYY_mn = np.vstack(
                (np.diag(YY),) * m
            )  # m*n matrix : each row is the diagonal y_i^Ty_i

            kXX = np.exp(
                -self.tuning_parameter * (dXX_mm + dXX_mm.transpose() - 2 * XX)
            )
            kYY = np.exp(
                -self.tuning_parameter * (dYY_nn + dYY_nn.transpose() - 2 * YY)
            )
            kXY = np.exp(-self.tuning_parameter * (dXX_mn + dYY_mn - 2 * XY))
        else:
            raise ValueError(f"Unknown kernel type: {self.kernel}")

        val = (
            (np.sum(kXX) - np.sum(np.diag(kXX))) / (m * (m - 1))
            + (np.sum(kYY) - np.sum(np.diag(kYY))) / (n * (n - 1))
            - 2 * np.sum(kXY) / (n * m)
        )
        if val < 0:
            val = 0

        return val

    def average(self, object_list: npt.NDArray) -> npt.NDArray:
        """Calculate the average of a list of distributions.

        Args:
            object_list (npt.NDArray[npt.NDArray]): List of distributions

        Returns:
            npt.NDArray: Average of the distributions
            (Returns is a mixture of vectors regardless of the kernel)

        """
        # filter out nan entries
        arrays_to_concatenate = [
            arr for arr in object_list if not np.any(np.isnan(arr))
        ]
        if len(arrays_to_concatenate) == 0:
            # NOTE: return nan distribution, since entries for DistributionKernelMMD
            # must have shape (?, d)
            return np.full((1, self.d), np.nan)
        return np.concatenate(arrays_to_concatenate, axis=0)


class DistributionWassersteinSamples(DataType):
    """Data type for distributions using Wasserstein distance
    where distributions are made with samples with the same number of samples.
    """

    def __init__(self, num_samples: int):
        """Initialize the distribution data type with Wasserstein distance.

        Args:
            num_samples (int): Number of samples in the distributions (n)

        """
        self.num_samples = num_samples

    def distance(self, obj1: npt.NDArray, obj2: npt.NDArray) -> float:
        """Calculate the Wasserstein distance between two distributions
        with the same number of samples.

        obj1 and obj2 should be 1-dimensional numpy arrays that represent
        empirical distributions.

        Args:
            obj1 (npt.NDArray): Distribution 1
            obj2 (npt.NDArray): Distribution 2

        Returns:
            float: Wasserstein distance between the two distributions

        """
        if len(obj1) != len(obj2):
            raise ValueError("Distributions must have the same number of samples")

        return float(np.mean(np.sum((np.sort(obj1) - np.sort(obj2)) ** 2)))

    def average(self, object_list: npt.NDArray[Any]) -> npt.NDArray:
        """Calculate the average of a list of distributions with the same
        number of samples.

        Args:
            object_list (npt.NDArray[Any]): List of distributions

        Returns:
            np.ndarray: Average of the distributions

        """
        # filter out nan values
        # All input objects should be 1-dimensional numpy arrays
        avg = np.nanmean([np.sort(obj) for obj in object_list], axis=0)
        if not isinstance(avg, float):
            # if avg is an array then return avg
            return avg
        else:
            # if avg is a float then it must (?) be a nan, so return an array of nans
            return np.full((self.num_samples,), np.nan)


class DistributionWassersteinQuantile(DataType):
    """Data type for distributions using Wasserstein distance
    where distributions are given by their quantile functions.
    """

    def empirical_quantile_function(
        self, samples: npt.NDArray
    ) -> Callable[[npt.NDArray], npt.NDArray]:
        """Create the quantile function of a distribution given samples.

        Args:
            samples (npt.NDArray): Samples of the distribution

        Returns:
            Callable[[npt.NDArray], npt.NDArray]: Quantile function of the distribution

        """
        samples_diff = np.concatenate(
            [np.array(samples[0]).reshape(1), np.diff(samples)]
        )

        def quantile_function(q: npt.NDArray) -> npt.NDArray:
            """Quantile function of the distribution.

            Args:
                q (npt.NDArray): Values between 0 and 1

            Returns:
                npt.NDArray: Quantile values

            """
            # Compute the empirical CDF values
            n = len(samples)
            cdf = np.arange(1, n + 1) / n
            # Use broadcasting to calculate the Heaviside contributions
            heaviside_matrix = np.heaviside(
                np.expand_dims(q, 1) - np.expand_dims(cdf, 0), 0.0
            )
            # Add a column of ones to the left of the Heaviside matrix
            first_col = np.ones(heaviside_matrix.shape[0]).reshape(-1, 1)
            heaviside_matrix = np.concatenate([first_col, heaviside_matrix], axis=1)
            # Remove the last column of Heaviside_matrix
            heaviside_matrix = heaviside_matrix[:, :-1]
            # Compute quantile values by summing contributions
            quantile_values = heaviside_matrix @ samples_diff

            return quantile_values

        return quantile_function

    def distance(
        self,
        obj1: Callable[[npt.NDArray], npt.NDArray],
        obj2: Callable[[npt.NDArray], npt.NDArray],
    ) -> float:
        """Calculate the Wasserstein distance between two distributions
        with the same number of samples.

        Args:
            obj1 (Callable[[npt.NDArray], npt.NDArray]): Distribution 1's quantile function
            obj2 (Callable[[npt.NDArray], npt.NDArray]): Distribution 2's quantile function

        Returns:
            float: Wasserstein distance between the two distributions

        """
        x = np.linspace(0, 1, 1000)
        return float(np.trapezoid((obj1(x) - obj2(x)) ** 2, x=x))

    def average(self, object_list: npt.NDArray) -> Callable:
        """Calculate the average of a list of quantile functions

        Args:
            object_list (npt.NDArray[Any]): List of quantile functions

        Returns:
            Callable: Average quantile function

        """
        # filter out float values
        object_list = np.array(
            [fn for fn in object_list if not isinstance(fn, float)], dtype=object
        )
        if len(object_list) == 0:
            return lambda x: x * 0

        def lin_comb_fn(quantiles: npt.NDArray) -> npt.NDArray:
            """Average a bunch of quantile functions.

            Args:
                quantiles (npt.NDArray): Values between 0 and 1

            Returns:
                npt.NDArray: Quantile values

            """
            # Compute the quantile values for each function
            quantile_values = np.stack([fn(quantiles) for fn in object_list])
            lin_comb_values = np.sum(quantile_values, axis=0) / len(object_list)
            return lin_comb_values

        return lin_comb_fn



"""MCAR simulation data

TODO: remove commented out code
"""

import numpy as np
from typing import Tuple


def gendata_lin_mcar(
    N: int, T: int, p: float, seed: int, r: int = 4
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Generates data using a bilinear model with uniform latent factors of dimension r.

    Parameters
    ----------
    N: int
        Number of users.
    T: int
        Number of time periods.
    p: float
        Probability of an entry being observed.
    seed: int
        Random seed for reproducibility.
    r: int, optional
        Dimension of latent factors. Default is 4.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray, np.ndarray]:
        Data (np.ndarray): Generated data matrix of shape (N, T).
        Theta (np.ndarray): Underlying true data matrix of shape (N, T).
        Masking (np.ndarray): Masking matrix indicating observed entries of shape (N, T).

    """
    # TODO: pass in random number generator rather than seed
    np.random.seed(seed=seed)
    ## Data Matrix (N * T)
    Data: np.ndarray = np.zeros((N, T))

    # user_range = 0.9
    # time_range = 0.9

    # user_std = 0.3
    # time_std = 0.3

    U = np.random.uniform(-1, 1, size=(N, r))  # * user_range * 2 - user_range
    V = np.random.uniform(-1, 1, size=(T, r))  # * time_range * 2 - time_range
    Y = np.matmul(U, V.transpose())
    # 1/np.sqrt(r) * np.matmul(U, V.transpose())

    # a = np.random.normal(0, user_std, size=N)
    # b = np.random.normal(0, time_std, size=T)
    # eps = np.random.normal(0, 0.05, size=(N, T))

    # treatment effect
    # delta(i,j) = a(i) + b(t) + eps(i,t)

    # aa = np.broadcast_to(a.reshape(N,1), (N,T))
    # bb = np.broadcast_to(b, (N,T))
    # delta = aa + bb + eps
    # Y1 = Y0 + delta

    # Y1 += np.random.normal(0, 0.001, size=(N,T))
    # gaussian noise
    Theta: np.ndarray = Y
    Y += np.random.normal(0, 0.001, size=(N, T))
    # TODO: clean up this code
    Masking: np.ndarray = np.zeros((N, T))

    Masking = np.reshape(np.random.binomial(1, p, (N * T)), (N, T))

    # Data[Masking == 1] = Y1[Masking == 1]
    # Data[Masking == 0] = Y0[Masking == 0]
    Data = np.array(Y)
    return Data, Theta, Masking


def gendata_nonlin_mcar(
    N: int, T: int, p: float, seed: int, non_lin: str = "expit", r: int = 4
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Generates data using a nonlinear model (default: expit) with uniform latent factors of dimension r.

    Parameters
    ----------
    N: int
        Number of users.
    T: int
        Number of time periods.
    p: float
        Probability of an entry being observed.
    seed: int
        Random seed for reproducibility.
    non_lin: str, optional
        Nonlinear function to apply. Options are "expit", "tanh", "sin", "cubic", "sinh". Default is "expit".
    r: int, optional
        Dimension of latent factors. Default is 4.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray, np.ndarray]:
        Data (np.ndarray): Generated data matrix of shape (N, T).
        Theta (np.ndarray): Latent variable matrix of shape (N, T).
        Masking (np.ndarray): Masking matrix indicating observed entries of shape (N, T).

    """
    # TODO: pass in random number generator rather than seed
    np.random.seed(seed=seed)

    def expit(x: np.ndarray) -> np.ndarray:
        return np.exp(x) / (1 + np.exp(x))

    Data = np.zeros((N, T))
    U = np.random.uniform(-1, 1, size=(N, r))
    V = np.random.uniform(-1, 1, size=(T, r))
    if non_lin == "expit":
        Y = expit(np.matmul(U, V.transpose()))
    elif non_lin == "tanh":
        Y = np.tanh(np.matmul(U, V.transpose()))
    elif non_lin == "sin":
        Y = np.sin(np.matmul(U, V.transpose()))
    elif non_lin == "cubic":
        Y = (np.matmul(U, V.transpose())) ** 3
    elif non_lin == "sinh":
        Y = np.sinh(np.matmul(U, V.transpose()))
    else:
        raise ValueError(
            "non_lin must be one of 'expit', 'tanh', 'sin', 'cubic', or 'sinh'."
        )
    Theta: np.ndarray = Y
    Y += np.random.normal(0, 0.001, size=(N, T))

    # TODO: clean up this code
    Masking: np.ndarray = np.zeros((N, T))

    Masking: np.ndarray = np.reshape(np.random.binomial(1, p, (N * T)), (N, T))

    # Data[Masking == 1] = Y1[Masking == 1]
    # Data[Masking == 0] = Y0[Masking == 0]
    Data: np.ndarray = np.array(Y)
    return Data, Theta, Masking



"""Tests for the AWNN imputer."""

import numpy as np
import pytest
from nsquared.aw_nn import aw_nn

# Initialize constants
ROWS = 16
COLS = 16


def test_impute_one_vs_all() -> None:
    """Test imputation of a single value vs. all values, ensure they align"""
    np.random.seed(0)  # For reproducibility
    data = np.random.randn(ROWS, COLS)
    mask = np.ones((ROWS, COLS))
    mask[1, 1] = 0  # Make one entry missing
    mask[2, 2] = 0  # Make another entry missing
    mask[3, 3] = 0  # Make another entry missing
    imputer = aw_nn()
    imputed_matrix = imputer.impute_all(data, mask)
    for i in range(ROWS):
        for j in range(COLS):
            if mask[i, j] == 0:
                imputed_value = imputer.impute(i, j, data, mask)
                assert np.isclose(imputed_value, imputed_matrix[i, j]), (
                    f"Imputed value at ({i}, {j}) does not match imputed matrix."
                )


def test_constant_imputation() -> None:
    """Test imputation on a matrix with constant values.
    The imputed value should be the same constant value.
    """
    data = np.ones((ROWS, COLS))
    mask = np.ones((ROWS, COLS))
    mask[1, 1] = 0  # Make one entry missing

    imputer = aw_nn(noise_variance=0.1, delta=0.05)

    imputed_value = imputer.impute(1, 1, data, mask)
    assert np.isclose(imputed_value, 1.0)


def test_no_observed_values() -> None:
    """Test imputation when there are no observed values in the column.
    Should return NaN.
    """
    data = np.ones((ROWS, COLS))
    mask = np.ones((ROWS, COLS))
    mask[:, 1] = 0  # Make entire column missing

    imputer = aw_nn()

    imputed_all = imputer.impute_all(data, mask)
    all_in_col_are_nan = np.isnan(imputed_all[:, 1])
    assert np.all(all_in_col_are_nan), "All imputed values in column should be NaN."


def test_weight_calculation() -> None:
    """Test that weights sum to approximately 1 and are non-negative."""
    data = np.array(
        [
            [1.0, 2.0, 3.0, 4.0],
            [1.1, 0.0, 3.1, 4.1],  # Second value will be imputed
            [1.2, 2.2, 3.2, 4.2],
            [1.3, 2.3, 3.3, 4.3],
        ]
    )
    mask = np.ones(data.shape)
    mask[1, 1] = 0

    imputer = aw_nn()

    # Get the weights through imputation
    imputed_value = imputer.impute(1, 1, data, mask)

    # The imputed value should be a weighted average of the observed values
    # in the same column, so it should be between min and max of those values
    observed_values = data[mask[:, 1] == 1, 1]
    assert observed_values.min() <= imputed_value <= observed_values.max()


def test_convergence() -> None:
    """Test that the fitting process converges within max_iterations."""
    np.random.seed(0)  # For reproducibility
    data = np.random.randn(ROWS, COLS)
    mask = np.ones((ROWS, COLS))
    # hide all diagonal and anti-diagonal elements to 0
    mask = mask - (np.eye(ROWS, COLS) * np.fliplr(np.eye(ROWS, COLS)))
    max_iterations = 100
    # Set a high max_iterations to ensure convergence
    imputer = aw_nn(
        noise_variance=0.1,
        delta=0.05,
        max_iterations=max_iterations,
    )
    # imputed_data = imputer.fit(data, mask, max_iterations=max_iterations)

    imputed_data = imputer.impute_all(data, mask)
    # Check that imputed data is not none and has the same shape as input
    assert imputed_data.shape == data.shape, (
        f"Shape mismatch: {imputed_data.shape} vs {data.shape}"
    )


def test_delta_parameter() -> None:
    """Test how different delta values affect the imputation."""
    np.random.seed(0)  # For reproducibility
    data = np.random.randn(ROWS, COLS)
    mask = np.ones((ROWS, COLS))
    mask[1, 1] = 0
    mask[2, 2] = 0

    # Compare imputation with different delta values
    imputer1 = aw_nn(noise_variance=0.1, delta=0.01)
    imputer2 = aw_nn(noise_variance=0.1, delta=0.5)

    value1 = imputer1.impute_all(data, mask)
    value2 = imputer2.impute_all(data, mask)
    assert not np.allclose(value1, value2), (
        "Imputed values should differ significantly with different delta values."
    )


def test_high_snr_convergence() -> None:
    """Test that with very high SNR, imputation result converges to the true signal.

    With a very small noise variance (high SNR), the imputed values should be
    very close to the true signal values.
    """
    np.random.seed(0)  # For reproducibility
    # Create a clean signal matrix with a simple pattern
    signal = np.zeros((ROWS, COLS))
    for i in range(ROWS):
        signal[i, :] = i + 1  # Distinct values for each row, e.g., 1, 2, 3, 4
    # Add extremely small noise (high SNR)
    noise_stddev = 1e-6  # Very small noise
    noisy_data = signal + np.random.normal(0, noise_stddev, (ROWS, COLS))
    # Create mask with a single missing value
    mask = np.ones((ROWS, COLS))
    mask[2, 2] = 0  # Make one entry missing

    # Create imputer with a very small noise variance
    imputer = aw_nn(
        noise_variance=noise_stddev**2,
        delta=0.05,
        max_iterations=100,
        convergence_threshold=1e-6,
    )

    # Impute the missing value
    # imputed_value = imputer.impute(2, 2, noisy_data, mask)
    imputed_all = imputer.impute_all(noisy_data, mask)
    # check if all imputed values are close to the true signal with some fault tolerance
    print(signal)
    # print(imputed_value)
    print(imputed_all)
    assert np.all(np.isclose(imputed_all, signal, atol=1e-3)), (
        "All imputed values should be close to the true signal values."
    )


def test_edge_case_single_observation() -> None:
    """Test imputation when there's only one observed value in the column."""
    data = np.ones((ROWS, COLS))
    mask = np.zeros((ROWS, COLS))
    mask[0, 1] = 1  # Only one observation in column 1

    imputer = aw_nn()
    # Should return the only observed value
    imputed_value = imputer.impute(0, 1, data, mask)
    assert np.isclose(imputed_value, data[0, 1])


def test_invalid_inputs() -> None:
    """Test that invalid inputs raise appropriate errors."""
    # Test with invalid delta values (should be between 0 and 1)
    with pytest.raises(ValueError, match="Delta must be between 0 and 1"):
        aw_nn(delta=-0.1)

    with pytest.raises(ValueError, match="Delta must be between 0 and 1"):
        aw_nn(delta=1.5)

    # Test with invalid noise variance (should be non-negative)
    with pytest.raises(ValueError, match="Noise variance must be non-negative"):
        aw_nn(noise_variance=-0.1)

    # Test with invalid convergence threshold (should be non-negative)
    with pytest.raises(ValueError, match="Convergence threshold must be non-negative"):
        aw_nn(convergence_threshold=-0.01)

    # Test with invalid max iterations (should be positive)
    with pytest.raises(ValueError, match="Max iterations must be positive"):
        aw_nn(max_iterations=0)

    with pytest.raises(ValueError, match="Max iterations must be positive"):
        aw_nn(max_iterations=-5)

    # Test with mismatched dimensions for data and mask
    imputer = aw_nn(noise_variance=0.1, delta=0.05)
    data = np.ones((3, 4))  # 3x4 matrix
    mask = np.ones((4, 3))  # 4x3 matrix (transposed dimensions)

    # Attempting to impute with mismatched dimensions should raise an error
    with pytest.raises(Exception):  # Could be IndexError, ValueError, etc.
        imputer.impute(1, 1, data, mask)

    # Test imputation with out-of-bounds indices
    data = np.ones((ROWS, COLS))
    mask = np.ones((ROWS, COLS))

    # Row index out of bounds
    with pytest.raises(Exception):  # Could be IndexError
        imputer.impute(ROWS + 5, 1, data, mask)

    # Column index out of bounds
    with pytest.raises(Exception):  # Could be IndexError
        imputer.impute(1, COLS + 5, data, mask)



import numpy as np
from nsquared.ts_nn import ts_nn


def test_full_neighborhood() -> None:
    """Test imputation when threshold is high so that all possible neighbors are used.

    For a 4x4 fully observed matrix with a missing cell at (2,2), the imputed value
    should be the average of the cross-product of all other rows and columns.
    """
    X = np.arange(16).reshape(4, 4).astype(float)
    A = np.ones((4, 4))
    # Make (2,2) missing.
    A[2, 2] = 0
    # With very high thresholds, neighbors are all rows (except row 2) and all columns (except col 2).
    imputer = ts_nn(
        distance_threshold_row=1e9, distance_threshold_col=1e9, is_percentile=False
    )
    estimated_value = imputer.impute(
        row=2, column=2, data_array=X, mask_array=A, allow_self_neighbor=False
    )
    # Neighbors: rows {0,1,3} and cols {0,1,3}
    vals = [
        X[0, 0],
        X[0, 1],
        X[0, 3],
        X[1, 0],
        X[1, 1],
        X[1, 3],
        X[3, 0],
        X[3, 1],
        X[3, 3],
    ]
    expected = np.mean(vals)
    assert np.isclose(estimated_value, expected), (
        f"Expected {expected}, got {estimated_value}"
    )


def test_single_neighbor() -> None:
    """Test imputation when there is exactly one neighbor.

    In a 2x2 matrix with one missing value at (1,1), if the only neighbor is row 0 and col 0,
    then the imputed value should equal X[0,0].
    """
    X = np.array([[5, 5], [5, 0]], dtype=float)
    A = np.array([[1, 1], [1, 0]])
    # With high thresholds, row neighbor of row 1 is {0} and column neighbor of col 1 is {0}.
    imputer = ts_nn(
        distance_threshold_row=1e9, distance_threshold_col=1e9, is_percentile=False
    )
    estimated_value = imputer.impute(
        row=1, column=1, data_array=X, mask_array=A, allow_self_neighbor=False
    )
    expected = X[0, 0]  # 5
    assert np.isclose(estimated_value, expected), (
        f"Expected {expected}, got {estimated_value}"
    )


def test_partial_neighborhood() -> None:
    """Test imputation on a matrix with some found neighbors.

    For a 3x3 matrix with a missing value at (1,1) and high thresholds,
    the imputed value should be the average of the cross-product of rows {0,2} and cols {0,2}.
    """
    X = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=float)
    A = np.ones_like(X)
    # Make (1,1) missing.
    A[1, 1] = 0
    imputer = ts_nn(
        distance_threshold_row=1e9, distance_threshold_col=1e9, is_percentile=False
    )
    estimated_value = imputer.impute(
        row=1, column=1, data_array=X, mask_array=A, allow_self_neighbor=False
    )
    # Neighbors: rows {0,2}, cols {0,2}
    vals = [X[0, 0], X[0, 2], X[2, 0], X[2, 2]]
    expected = np.mean(vals)
    assert np.isclose(estimated_value, expected), (
        f"Expected {expected}, got {estimated_value}"
    )


def test_no_neighbors_fallback() -> None:
    """Test fallback when no neighbor is found.

    With threshold zero, no neighbor is selected. Then if the target cell is missing,
    the imputation should return np.nan
    """
    X = np.array([[20, 20], [30, 40]])
    A = np.array([[0, 1], [1, 1]])
    imputer = ts_nn(
        distance_threshold_row=0, distance_threshold_col=0, is_percentile=False
    )
    estimated_value = imputer.impute(
        row=0, column=0, data_array=X, mask_array=A, allow_self_neighbor=False
    )
    assert np.isnan(estimated_value), f"Expected NaN, got {estimated_value}"


def test_multiple_missing_in_neighborhood() -> None:
    """Test imputation when some neighbor entries are also missing.

    In a 3x3 matrix with target missing at (1,1) and an additional missing value at (0,0)
    in the neighborhood, the average is computed only over the observed values.
    """
    X = np.array([[10, 20, 30], [40, 50, 60], [70, 80, 90]], dtype=float)
    A = np.ones_like(X)
    # Make target (1,1) missing and also (0,0) missing.
    A[1, 1] = 0
    A[0, 0] = 0
    imputer = ts_nn(
        distance_threshold_row=1e9, distance_threshold_col=1e9, is_percentile=False
    )
    estimated_value = imputer.impute(
        row=1, column=1, data_array=X, mask_array=A, allow_self_neighbor=False
    )
    # Neighbors: rows {0,2}, cols {0,2} → cross product = {(0,0), (0,2), (2,0), (2,2)}
    # (0,0) is missing, so use (0,2), (2,0), (2,2)
    vals = [X[0, 2], X[2, 0], X[2, 2]]
    expected = np.mean(vals)
    assert np.isclose(estimated_value, expected), (
        f"Expected {expected}, got {estimated_value}"
    )


def test_imputation_stability() -> None:
    """Test that repeated imputation on the same input returns the same value.

    This ensures that the TSNN imputation is deterministic.
    """
    X = np.random.rand(5, 5)
    A = np.ones((5, 5))
    # Make target (2,2) missing.
    A[2, 2] = 0
    imputer = ts_nn(
        distance_threshold_row=1e9, distance_threshold_col=1e9, is_percentile=False
    )
    val1 = imputer.impute(
        row=2, column=2, data_array=X, mask_array=A, allow_self_neighbor=False
    )
    val2 = imputer.impute(
        row=2, column=2, data_array=X, mask_array=A, allow_self_neighbor=False
    )
    print(val1)
    assert np.isclose(val1, val2), f"Imputation not stable: {val1} vs {val2}"


def test_random_noise_imputation() -> None:
    """Test that imputation on a random matrix with missing entries does not return NaN.

    This checks that the TSNN imputer can handle randomness and produce a finite result.
    """
    np.random.seed(0)
    X = np.random.rand(6, 6)
    A = np.ones((6, 6))
    # Introduce missingness at a few random locations.
    A[1, 3] = 0
    A[4, 2] = 0
    imputer = ts_nn(
        distance_threshold_row=1e9, distance_threshold_col=1e9, is_percentile=False
    )
    estimated_value = imputer.impute(
        row=1, column=3, data_array=X, mask_array=A, allow_self_neighbor=False
    )
    assert np.isfinite(estimated_value), (
        f"Imputed value is not finite: {estimated_value}"
    )


def test_edge_case_empty_matrix() -> None:
    """Test imputation on an empty matrix (no rows or columns)."""
    X = np.empty((0, 0))
    A = np.empty((0, 0))
    imputer = ts_nn(
        distance_threshold_row=1e9, distance_threshold_col=1e9, is_percentile=False
    )
    try:
        imputer.impute(
            row=0, column=0, data_array=X, mask_array=A, allow_self_neighbor=False
        )
    except IndexError:
        pass  # Expect an IndexError due to the empty matrix
    else:
        raise AssertionError("Expected IndexError for empty matrix.")


def test_single_row_missing() -> None:
    """Test imputation with a matrix where an entire row is missing."""
    X = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=float)
    A = np.ones_like(X)
    # Make row 1 entirely missing.
    A[1, :] = 0
    imputer = ts_nn(
        distance_threshold_row=1e9, distance_threshold_col=1e9, is_percentile=False
    )
    estimated_value = imputer.impute(
        row=1, column=1, data_array=X, mask_array=A, allow_self_neighbor=False
    )
    # nan because N_row is the empty set, and the cross product of an empty set with any set is empty.
    assert np.isnan(estimated_value), f"Expected NaN, got {estimated_value}"


def test_single_column_missing() -> None:
    """Test imputation with a matrix where an entire column is missing."""
    X = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=float)
    A = np.ones_like(X)
    # Make column 1 entirely missing.
    A[:, 1] = 0
    imputer = ts_nn(
        distance_threshold_row=1e9, distance_threshold_col=1e9, is_percentile=False
    )
    estimated_value = imputer.impute(
        row=1, column=1, data_array=X, mask_array=A, allow_self_neighbor=False
    )
    # nan because N_col is the empty set, and the cross product of any set with an empty set is empty.
    assert np.isnan(estimated_value), f"Expected NaN, got {estimated_value}"


def test_multiple_neighbors_with_finite_thresholds() -> None:
    """Test imputation with multiple neighbors but finite distance thresholds."""
    X = np.array([[1, 2, 30], [4, 5, 6], [7, 8, 9]], dtype=float)
    A = np.ones_like(X)
    # Make (1,1) missing, and rows 0 and 2 should be considered as neighbors.
    A[1, 1] = 0
    imputer = ts_nn(
        distance_threshold_row=1000, distance_threshold_col=10000, is_percentile=False
    )
    estimated_value = imputer.impute(
        row=1, column=1, data_array=X, mask_array=A, allow_self_neighbor=False
    )
    # Neighbors: rows {0, 2}, cols {0, 2}
    vals = [X[0, 0], X[0, 2], X[2, 0], X[2, 2]]
    expected = np.mean(vals)
    assert np.isclose(estimated_value, expected), (
        f"Expected {expected}, got {estimated_value}"
    )


def test_find_one_neighbor() -> None:
    """Test imputation with a singular near-infinite neighbor but finite distance thresholds."""
    X = np.array([[1, 2, 30], [4, 5, 6], [7, 8, 9e99]], dtype=float)
    A = np.ones_like(X)
    A[1, 1] = 0
    imputer = ts_nn(
        distance_threshold_row=1000, distance_threshold_col=1000, is_percentile=False
    )
    estimated_value = imputer.impute(
        row=1, column=1, data_array=X, mask_array=A, allow_self_neighbor=False
    )
    # Neighbors: rows {0}, cols {0}
    vals = [X[0, 0]]
    expected = np.mean(vals)
    assert np.isclose(estimated_value, expected), (
        f"Expected {expected}, got {estimated_value}"
    )


def test_low_row_high_col_threshold() -> None:
    """Test imputation with a low row threshold and a high column threshold.

    In this case, only row 0 is selected as the neighbor, but all columns {0, 1, 2}
    are considered as neighbors due to the high column threshold.
    """
    X = np.array([[1, 2, 3], [10, 20, 30], [10000, 20000, 30000]], dtype=float)
    A = np.ones_like(X)
    # Make (1,1) missing.
    A[1, 1] = 0
    imputer = ts_nn(
        distance_threshold_row=1000, distance_threshold_col=1e9, is_percentile=False
    )
    estimated_value = imputer.impute(
        row=1, column=1, data_array=X, mask_array=A, allow_self_neighbor=False
    )
    # Neighbors: row {0}, cols {0, 1, 2} → cross product = {(0,0), (0,1), (0,2)}
    vals = [X[0, 0], X[0, 1], X[0, 2]]
    expected = np.mean(vals)
    assert np.isclose(estimated_value, expected), (
        f"Expected {expected}, got {estimated_value}"
    )


def test_low_col_high_row_threshold() -> None:
    """Test imputation with a low column threshold and a high row threshold.

    In this case, only column 0 is selected as the neighbor, but all rows {0, 1, 2}
    are considered as neighbors due to the high column threshold.
    """
    X = np.array([[1, 2, 3], [4, 5, 6], [7, 20000, 30000]], dtype=float)
    A = np.ones_like(X)
    # Make (1,1) missing.
    A[1, 1] = 0
    imputer = ts_nn(
        distance_threshold_row=1000, distance_threshold_col=1e9, is_percentile=False
    )
    estimated_value = imputer.impute(
        row=1, column=1, data_array=X, mask_array=A, allow_self_neighbor=False
    )
    # Neighbors: row {0}, cols {0, 1, 2} → cross product = {(0,0), (0,1), (0,2)}
    vals = [X[0, 0], X[0, 1], X[0, 2]]
    expected = np.mean(vals)
    assert np.isclose(estimated_value, expected), (
        f"Expected {expected}, got {estimated_value}"
    )



# TODO these should be maintained as this __init__.py file is used to expose the classes and functions when the package itself is imported. Task owner: Aashish
# For example: `from nsquared import NearestNeighborImputer`
# is easier to read than from `nsquared.nnimputer import NearestNeighborImputer`

# TODO @ALL: please uncomment the following lines when the code is ready in each file

from .dr_nn import *  # noqa: F403

# from .nadaraya_watson import * # noqa: F403
from .nnimputer import *  # noqa: F403

# from .syn_nn import * # noqa: F403
# from .ts_nn import * # noqa: F403
from .vanilla_nn import *  # noqa: F403
from .utils import *  # noqa: F403
from .simulations import *  # noqa: F403
# ...add new files here

from .datasets.dataloader_factory import *  # noqa: F403
from .datasets.dataloader_base import *  # noqa: F403



"""Base class for all nearest neighbors algorithms."""

import numpy.typing as npt
from abc import ABC, abstractmethod
from typing import Any, Union, Tuple
from hyperopt import Trials


class DataType(ABC):
    """Abstract class for data types. Examples include scalars and distributions."""

    @abstractmethod
    def distance(self, obj1: Any, obj2: Any) -> float:
        """Calculate the distance between two objects.

        Args:
            obj1 (Any): Object 1
            obj2 (Any): Object 2

        """
        pass

    @abstractmethod
    def average(self, object_list: npt.NDArray[Any]) -> Any:
        """Calculate the average of a list of objects.

        Args:
            object_list (npt.NDArray[Any]): List of objects

        """
        pass


class EstimationMethod(ABC):
    """Abstract class for estimation methods.
    Examples include row-row, col-col, two-sided, and doubly-robust.
    """

    def __init__(self, is_percentile: bool = True):
        """Initialize the estimation method.

        Args:
            is_percentile (bool): Whether to use percentile-based threshold. Defaults to True.

        """
        self.is_percentile = is_percentile

    @abstractmethod
    def impute(
        self,
        row: int,
        column: int,
        data_array: npt.NDArray,
        mask_array: npt.NDArray,
        distance_threshold: Union[float, Tuple[float, float]],
        data_type: DataType,
        allow_self_neighbor: bool = False,
        **kwargs: Any,
    ) -> npt.NDArray:
        """Impute the missing value at the given row and column.

        Args:
            row (int): Row index
            column (int): Column index
            data_array (npt.NDArray): Data matrix
            mask_array (npt.NDArray): Mask matrix
            distance_threshold (float): Distance threshold for nearest neighbors
            data_type (DataType): Data type to use (e.g. scalars, distributions)
            allow_self_neighbor (bool): Whether to allow self-neighbor. Defaults to False.
            **kwargs (Any): Additional arguments for the imputer

        Returns:
            npt.NDArray: Imputed value

        """
        pass

    @abstractmethod
    def _calculate_distances(
        self,
        row: int,
        col: int,
        data_array: npt.NDArray,
        mask_array: npt.NDArray,
        data_type: DataType,
    ) -> None:
        """Sets the distances for the imputer.
        Sets the distances as a class attribute, so returns nothing.

        Args:
            row (int): Row index
            col (int): Column index
            data_array (npt.NDArray): Data matrix
            mask_array (npt.NDArray): Mask matrix
            data_type (DataType): Data type to use (e.g. scalars, distributions)

        """
        pass

    def impute_all(
        self,
        data_array: npt.NDArray,
        mask_array: npt.NDArray,
        distance_threshold: Union[float, Tuple[float, float]],
        data_type: DataType,
    ) -> npt.NDArray:
        """Impute all missing values in the data array.
        Note that this is not an abstract method, but a default implementation.
        This method can be overridden by subclasses if needed.

        Args:
            data_array (npt.NDArray): Data matrix
            mask_array (npt.NDArray): Mask matrix
            distance_threshold (float): Distance threshold for nearest neighbors
            data_type (DataType): Data type to use (e.g. scalars, distributions)

        Returns:
            npt.NDArray: Imputed value

        """
        # by default, just call impute for each missing value
        n_rows, n_cols = data_array.shape
        imputed_data = data_array.copy()
        for i in range(n_rows):
            for j in range(n_cols):
                imputed_data[i, j] = self.impute(
                    i,
                    j,
                    data_array,
                    mask_array,
                    distance_threshold,
                    data_type,
                )
        return imputed_data


class NearestNeighborImputer:
    """Nearest neighbor composed of different kinds of methods."""

    def __init__(
        self,
        estimation_method: EstimationMethod,
        data_type: DataType,
        distance_threshold: Union[float, Tuple[float, float], None] = None,
    ):
        """Initialize the imputer.

        Args:
            estimation_method (EstimationMethod): Estimation method to use (e.g. row-row, col-col, two-sided, doubly-robust)
            data_type (DataType): Data type to use (e.g. scalars, distributions)
            distance_threshold (Optional[float], Optional[Tuple[float, float]] optional): Distance threshold(s) to use. Defaults to None.

        """
        self.estimation_method = estimation_method
        self.data_type = data_type
        self.distance_threshold = distance_threshold

    def __str__(self):
        return f"NearestNeighborImputer(estimation_method={self.estimation_method}, data_type={self.data_type})"

    def impute(
        self,
        row: int,
        column: int,
        data_array: npt.NDArray,
        mask_array: npt.NDArray,
        **kwargs: Any,
    ) -> npt.NDArray:
        """Impute the missing value at the given row and column.

        Args:
            row (int): Row index
            column (int): Column index
            data_array (npt.NDArray): Data matrix
            mask_array (npt.NDArray): Mask matrix
            **kwargs (Any): Additional keyword arguments

        Raises:
            ValueError: If distance threshold is not set

        Returns:
            npt.NDArray: Imputed value

        """
        if self.distance_threshold is None:
            raise ValueError(
                "Distance threshold is not set. Call a FitMethod on this imputer or manually set it."
            )
        return self.estimation_method.impute(
            row,
            column,
            data_array,
            mask_array,
            self.distance_threshold,
            self.data_type,
            **kwargs,
        )

    def impute_all(
        self,
        data_array: npt.NDArray,
        mask_array: npt.NDArray,
    ) -> npt.NDArray:
        """Impute all missing values in the data array.

        Args:
            data_array (npt.NDArray): Data matrix
            mask_array (npt.NDArray): Mask matrix
            distance_threshold (float): Distance threshold for nearest neighbors

        Returns:
            npt.NDArray: Imputed value

        """
        if self.distance_threshold is None:
            raise ValueError(
                "Distance threshold is not set. Call a FitMethod on this imputer or manually set it."
            )
        return self.estimation_method.impute_all(
            data_array, mask_array, self.distance_threshold, self.data_type
        )


class FitMethod(ABC):
    """Abstract class for fitting methods.
    Examples include cross validation methods.
    """

    @abstractmethod
    def fit(
        self,
        data_array: npt.NDArray,
        mask_array: npt.NDArray,
        imputer: NearestNeighborImputer,
        ret_trials: bool = False,
    ) -> Union[
        float,
        Tuple[float, float],
        Tuple[float, Trials],
        Tuple[Tuple[float, float], Trials],
    ]:
        """Find the best distance threshold for the given data.

        Args:
            data_array (npt.NDArray): Data matrix
            mask_array (npt.NDArray): Mask matrix
            imputer (NearestNeighborImputer): Imputer object
            ret_trials (bool): Whether to return metadata about the hyperparameter search. Defaults to False.

        Returns:
            float | Tuple[float, float]: Best distance threshold(s)

        """
        pass



# Enable LaTeX for text rendering
import matplotlib as mpl

mpl.rcParams["text.latex.preamble"] = r"\usepackage{amssymb} \usepackage{amsmath}"
mpl.rcParams["text.usetex"] = True
mpl.rcParams["font.family"] = "serif"


# Single column figures
TICK_FONT_SIZE = 6
MINOR_TICK_FONT_SIZE = 3
LABEL_FONT_SIZE = 10
LEGEND_FONT_SIZE = 7
MARKER_ALPHA = 1
MARKER_SIZE = 3
LINE_ALPHA = 0.75
OUTWARD = 4

METHOD_ALIASES = {
    "kernel": "Kernel-\nNN",
    "wasserstein_samples": "W$_2$-\nNN",
    "auto": "Auto-\nNN",
    "row-row": "Row-\nNN",
    "col-col": "Col-\nNN",
    "dr": "DR-\nNN",
    "ts": "TS-\nNN",
    "usvt": "USVT",
    "softimpute": "Soft\nImpute",
    "nadaraya": "NW",
    "aw": "AW-\nNN",
    "sc": "SC",
}

METHOD_ALIASES_SINGLE_LINE = {
    "auto": "AutoNN",
    "row-row": "RowNN",
    "col-col": "ColNN",
    "dr": "DRNN",
    "ts": "TSNN",
    "usvt": "USVT",
    "softimpute": "SI",
    "nadaraya": "NW",
    "aw": "AWNN",
    "sc": "SC",
    "kernel": "KernelNN",
    "wasserstein_samples": "W$_2$NN",
}

METHOD_LINE_STYLES = {
    "row-row": "--",
    "col-col": "--",
    "dr": "dotted",
    "ts": "dotted",
    "usvt": "-",
    "softimpute": "-",
    "nadaraya": "-",
    "aw": "dashdot",
    "sc": "-",
    "auto": "-",
}

NEURIPS_TEXTWIDTH = 5.5

COLORS = {
    "row-row": "red",
    "auto": "grey",
    "kernel": "dimgrey",
    "wasserstein_samples": "dimgrey",
    # "row-row": "grey",
    "col-col": "grey",
    "dr": "grey",
    "ts": "grey",
    "usvt": "lightgrey",
    "softimpute": "lightgrey",
    "aw": "grey",
    "sc": "blue",
}

DATA_TYPE_ALIASES = {
    "kernel_mmd": "MMD",
    "wasserstein_samples": "W$_2$",
}



from .nnimputer import FitMethod, DataType, NearestNeighborImputer
from .estimation_methods import DREstimator, TSEstimator, AutoEstimator
import numpy.typing as npt
from hyperopt import hp, fmin, tpe, Trials
from typing import cast, Union, Any
import numpy as np


def evaluate_imputation(
    data_array: npt.NDArray,
    mask_array: npt.NDArray,
    imputer: NearestNeighborImputer,
    test_cells: list[tuple[int, int]],
    data_type: DataType,
    allow_self_neighbor: bool = False,
    **kwargs: Any,
) -> float:
    """Evaluate the imputer on a set.

    Args:
        data_array (npt.NDArray): Data matrix
        mask_array (npt.NDArray): Mask matrix
        imputer (NearestNeighborImputer): Imputer object
        test_cells (list[tuple[int,int]]): List of cells as tuples of row and column indices
        data_type (DataType): Data type to use (e.g. scalars, distributions)
        allow_self_neighbor (bool, optional): Whether to allow the entry itself as a neighbor. Defaults to False.
        **kwargs (Any): Additional keyword args

    Raises:
        ValueError: If a validation cell is missing

    Returns:
        float: Average imputation error

    """
    # Block out the test cells
    for row, col in test_cells:
        if mask_array[row, col] == 0 or data_array[row, col] is None:
            raise ValueError("Validation cell is missing.")
        mask_array[row, col] = 0  # Set the mask to missing
    errors = []
    for row, col in test_cells:
        imputed_value = imputer.impute(
            row, col, data_array, mask_array, allow_self_neighbor=allow_self_neighbor
        )
        true_value = data_array[row, col]
        errors.append(data_type.distance(imputed_value, true_value))

    # Reset the mask
    for row, col in test_cells:
        mask_array[row, col] = 1

    return float(np.nanmean(errors))


class LeaveBlockOutValidation(FitMethod):
    """Fit method by leaving out a block of cells."""

    def __init__(
        self,
        block: list[tuple[int, int]],
        distance_threshold_range: tuple[float, float],
        n_trials: int,
        data_type: DataType,
        allow_self_neighbor: bool = False,
        rng: np.random.Generator | None = None,
    ):
        """Initialize the block fit method.

        Args:
            block (list[tuple[int,int]]): List of cells as tuples of row and column indices
            distance_threshold_range (tuple[float,float]): Range of distance thresholds to test
            n_trials (int): Number of trials to run
            data_type (DataType): Data type to use (e.g. scalars, distributions)
            allow_self_neighbor (bool, optional): Whether to allow the entry itself as a neighbor. Defaults to False.
            rng (np.random.Generator | None, optional): Random number generator. Defaults to None.

        """
        self.block = block
        self.distance_threshold_range = distance_threshold_range
        self.n_trials = n_trials
        self.data_type = data_type
        self.allow_self_neighbor = allow_self_neighbor
        self.rng = rng

    def fit(
        self,
        data_array: npt.NDArray,
        mask_array: npt.NDArray,
        imputer: NearestNeighborImputer,
        ret_trials: bool = False,
        verbose: bool = False,
    ) -> Union[float, tuple[float, Trials]]:
        """Find the best distance threshold for the given data
        by leaving out a block of cells and testing imputation against them.

        Args:
            data_array (npt.NDArray): Data matrix
            mask_array (npt.NDArray): Mask matrix
            imputer (NearestNeighborImputer): Imputer object
            ret_trials (bool): If True, return the trials object which contains metadata on hyperparameter search.
            verbose (bool, optional): Whether to print the progress. Defaults to False.

        Returns:
            float: Best distance threshold or (float, Trials): Best distance threshold and trials object if ret_trials is True.

        """

        def objective(distance_threshold: float) -> float:
            """Objective function for hyperopt.

            Args:
                distance_threshold (float): Distance threshold to test

            Returns:
                float: Average imputation error

            """
            imputer.distance_threshold = distance_threshold
            return evaluate_imputation(
                data_array,
                mask_array,
                imputer,
                self.block,
                self.data_type,
                self.allow_self_neighbor,
            )

        lower_bound, upper_bound = self.distance_threshold_range
        trials = Trials()
        best_distance_threshold = fmin(
            fn=objective,
            verbose=verbose,
            space=hp.uniform("distance_threshold", lower_bound, upper_bound),
            algo=tpe.suggest,
            max_evals=self.n_trials,
            trials=trials,
            rstate=self.rng,
        )
        if best_distance_threshold is None:
            return float("nan")
        imputer.distance_threshold = best_distance_threshold["distance_threshold"]

        if ret_trials:
            return imputer.distance_threshold, trials
        return imputer.distance_threshold


class DualThresholdLeaveBlockOutValidation(FitMethod):
    """An abstract base subclass for fit methods that leave out a block of cells and tune separate row and column thresholds."""

    expected_estimator_type = type(None)

    def __init__(
        self,
        block: list[tuple[int, int]],
        distance_threshold_range_row: tuple[float, float],
        distance_threshold_range_col: tuple[float, float],
        n_trials: int,
        data_type: DataType,
        allow_self_neighbor: bool = False,
    ):
        """Initialize the dual threshold block fit method.

        Args:
            block (list[tuple[int, int]]): List of cells as tuples of row and column indices.
            distance_threshold_range_row (tuple[float, float]): Range of row distance thresholds to test.
            distance_threshold_range_col (tuple[float, float]): Range of column distance thresholds to test.
            n_trials (int): Number of trials to run.
            data_type (DataType): Data type to use (e.g. scalars, distributions).
            allow_self_neighbor (bool, optional): Whether to allow the entry itself as a neighbor. Defaults to False.

        """
        self.block = block
        self.distance_threshold_range_row = distance_threshold_range_row
        self.distance_threshold_range_col = distance_threshold_range_col
        self.n_trials = n_trials
        self.data_type = data_type
        self.allow_self_neighbor = allow_self_neighbor

    def fit(
        self,
        data_array: npt.NDArray,
        mask_array: npt.NDArray,
        imputer: NearestNeighborImputer,
        ret_trials: bool = False,
    ) -> Union[tuple[float, float], tuple[tuple[float, float], Trials]]:
        """Find the best distance thresholds for rows and columns by leaving out a block of cells and testing imputation.

        Args:
            data_array (npt.NDArray): Data matrix.
            mask_array (npt.NDArray): Mask matrix.
            imputer (NearestNeighborImputer): Imputer object.
            ret_trials (bool): If True, return the trials object which contains metadata on hyperparameter search.

        Returns:
            tuple[float, float]: Best distance thresholds for rows and columns.

        """

        def _objective(params: dict[str, float]) -> float:
            """Objective function for hyperopt.

            Args:
                params (dict[str, float]): Dictionary containing row and column distance thresholds

            Returns:
                float: Average imputation error

            """
            row_threshold = params["distance_threshold_row"]
            col_threshold = params["distance_threshold_col"]
            imputer.distance_threshold = (row_threshold, col_threshold)
            return evaluate_imputation(
                data_array,
                mask_array,
                imputer,
                self.block,
                self.data_type,
                self.allow_self_neighbor,
            )

        lower_bound_row, upper_bound_row = self.distance_threshold_range_row
        lower_bound_col, upper_bound_col = self.distance_threshold_range_col
        trials = Trials()
        best_params = fmin(
            fn=_objective,
            space={
                "distance_threshold_row": hp.uniform(
                    "distance_threshold_row", lower_bound_row, upper_bound_row
                ),
                "distance_threshold_col": hp.uniform(
                    "distance_threshold_col", lower_bound_col, upper_bound_col
                ),
            },
            algo=tpe.suggest,
            max_evals=self.n_trials,
            verbose=False,
            trials=trials,
        )

        if best_params is None:
            return float("nan"), float("nan")

        imputer.distance_threshold = (
            best_params["distance_threshold_row"],
            best_params["distance_threshold_col"],
        )
        if ret_trials:
            return imputer.distance_threshold, trials
        return imputer.distance_threshold


class DRLeaveBlockOutValidation(DualThresholdLeaveBlockOutValidation):
    """Fit method by leaving out a block of cells using separate thresholds for rows and columns with a DREstimator."""

    expected_estimator_type = DREstimator

    def fit(
        self,
        data_array: npt.NDArray,
        mask_array: npt.NDArray,
        imputer: NearestNeighborImputer,
        ret_trials: bool = False,
    ) -> Union[tuple[float, float], tuple[tuple[float, float], Trials]]:
        """Find the best distance thresholds for rows and columns using a DREstimator.

        Args:
            data_array (npt.NDArray): Data matrix.
            mask_array (npt.NDArray): Mask matrix.
            imputer (NearestNeighborImputer): Imputer object.
            ret_trials (bool): If True, return the trials object which contains metadata on hyperparameter search.

        Returns:
            tuple[float, float]: Best distance thresholds for rows and columns.

        Raises:
            ValueError: If the imputer does not use a DREstimator.

        """
        if not isinstance(imputer.estimation_method, DREstimator):
            raise ValueError(
                f"The imputer must use a DREstimator for {self.__class__.__name__}."
            )
        imputer.estimation_method = cast(DREstimator, imputer.estimation_method)
        return super().fit(data_array, mask_array, imputer, ret_trials)


class TSLeaveBlockOutValidation(DualThresholdLeaveBlockOutValidation):
    """Fit method by leaving out a block of cells using separate thresholds for rows and columns with a TSEstimator."""

    expected_estimator_type = TSEstimator

    def fit(
        self,
        data_array: npt.NDArray,
        mask_array: npt.NDArray,
        imputer: NearestNeighborImputer,
        ret_trials: bool = False,
    ) -> Union[tuple[float, float], tuple[tuple[float, float], Trials]]:
        """Find the best distance thresholds for rows and columns using a TSEstimator.

        Args:
            data_array (npt.NDArray): Data matrix.
            mask_array (npt.NDArray): Mask matrix.
            imputer (NearestNeighborImputer): Imputer object.
            ret_trials (bool): If True, return the trials object which contains metadata on hyperparameter search.

        Returns:
            tuple[float, float]: Best distance thresholds for rows and columns.

        Raises:
            ValueError: If the imputer does not use a TSEstimator.

        """
        if not isinstance(imputer.estimation_method, TSEstimator):
            raise ValueError(
                f"The imputer must use a TSEstimator for {self.__class__.__name__}."
            )
        imputer.estimation_method = cast(TSEstimator, imputer.estimation_method)
        return super().fit(data_array, mask_array, imputer)


class AutoDRTSLeaveBlockOutValidation(DualThresholdLeaveBlockOutValidation):
    """Fit method by leaving out a block of cells using separate thresholds for rows and columns with a AutoEstimator."""

    expected_estimator_type = AutoEstimator

    def __init__(
        self,
        block: list[tuple[int, int]],
        distance_threshold_range_row: tuple[float, float],
        distance_threshold_range_col: tuple[float, float],
        alpha_range: tuple[float, float],
        n_trials: int,
        data_type: DataType,
        allow_self_neighbor: bool = False,
    ):
        """Initialize the dual threshold block fit method with AutoEstimator.

        Args:
            block (list[tuple[int, int]]): List of cells as tuples of row and column indices.
            distance_threshold_range_row (tuple[float, float]): Range of row distance thresholds to test.
            distance_threshold_range_col (tuple[float, float]): Range of column distance thresholds to test.
            alpha_range (tuple[float, float]): Range of alpha values to test.
            n_trials (int): Number of trials to run.
            data_type (DataType): Data type to use (e.g. scalars, distributions).
            allow_self_neighbor (bool, optional): Whether to allow the entry itself as a neighbor. Defaults to False.

        """
        self.alpha_range = alpha_range
        super().__init__(
            block,
            distance_threshold_range_row,
            distance_threshold_range_col,
            n_trials,
            data_type,
        )
        self.allow_self_neighbor = allow_self_neighbor

    def fit(
        self,
        data_array: npt.NDArray,
        mask_array: npt.NDArray,
        imputer: NearestNeighborImputer,
        ret_trials: bool = False,
    ) -> Union[tuple[float, float], tuple[tuple[float, float], Trials]]:
        """Find the best distance thresholds for rows and columns by leaving out a block of cells and testing imputation.

        Args:
            data_array (npt.NDArray): Data matrix.
            mask_array (npt.NDArray): Mask matrix.
            imputer (NearestNeighborImputer): Imputer object.
            ret_trials (bool): If True, return the trials object which contains metadata on hyperparameter search.

        Returns:
            tuple[float, float]: Best distance thresholds for rows and columns.

        """
        if not isinstance(imputer.estimation_method, AutoEstimator):
            raise ValueError(
                f"The imputer must use a AutoEstimator for {self.__class__.__name__}."
            )
        imputer.estimation_method = cast(AutoEstimator, imputer.estimation_method)

        def _objective(params: dict[str, float]) -> float:
            """Objective function for hyperopt.

            Args:
                params (dict[str, float]): Dictionary containing row and column distance thresholds

            Returns:
                float: Average imputation error

            """
            row_threshold = params["distance_threshold_row"]
            col_threshold = params["distance_threshold_col"]
            alpha = params["alpha"]
            imputer.distance_threshold = (row_threshold, col_threshold)
            if not isinstance(imputer.estimation_method, AutoEstimator):
                raise ValueError(
                    f"The imputer must use a AutoEstimator for {self.__class__.__name__}."
                )
            imputer.estimation_method.alpha = alpha
            return evaluate_imputation(
                data_array,
                mask_array,
                imputer,
                self.block,
                self.data_type,
                self.allow_self_neighbor,
            )

        lower_bound_row, upper_bound_row = self.distance_threshold_range_row
        lower_bound_col, upper_bound_col = self.distance_threshold_range_col
        lower_bound_alpha, upper_bound_alpha = self.alpha_range
        trials = Trials()
        best_params = fmin(
            fn=_objective,
            space={
                "distance_threshold_row": hp.uniform(
                    "distance_threshold_row", lower_bound_row, upper_bound_row
                ),
                "distance_threshold_col": hp.uniform(
                    "distance_threshold_col", lower_bound_col, upper_bound_col
                ),
                "alpha": hp.uniform("alpha", lower_bound_alpha, upper_bound_alpha),
            },
            algo=tpe.suggest,
            max_evals=self.n_trials,
            verbose=False,
            trials=trials,
        )

        if best_params is None:
            return float("nan"), float("nan")

        imputer.distance_threshold = (
            best_params["distance_threshold_row"],
            best_params["distance_threshold_col"],
        )
        imputer.estimation_method.alpha = best_params["alpha"]

        if ret_trials:
            return imputer.distance_threshold, trials
        return imputer.distance_threshold



"""A function to return the TSNN imputer.

More details on the Two Sided Nearest Neighbor (TSNN) method can be found in Section 3 of the following paper:
    Sadhukhan, T., Paul, M., Dwivedi, R. (2024).
    On adaptivity and minimax optimality of two-sided nearest neighbors. arXiv preprint arXiv:2411.12965.
"""

from .nnimputer import NearestNeighborImputer
from typing import Optional


def ts_nn(
    distance_threshold_row: Optional[float] = None,
    distance_threshold_col: Optional[float] = None,
    is_percentile: bool = True,
) -> NearestNeighborImputer:
    """Create a two-sided nearest neighbor imputer.

    If distance_threshold_row and distance_threshold_col are not provided, they must be set by calling fit on the imputer.

    Args:
        distance_threshold_row (float): [Optional] Distance threshold for row-row nearest neighbors.
        distance_threshold_col (float): [Optional] Distance threshold for column-column nearest neighbors.
        is_percentile (bool): [Optional] Whether to use percentile-based distance threshold. Defaults to False.

    Returns:
        NearestNeighborImputer: A two-sided nearest neighbor imputer.

    """
    from .estimation_methods import TSEstimator
    from .data_types import Scalar

    estimator = TSEstimator(is_percentile=is_percentile)
    data_type = Scalar()
    if distance_threshold_row is None or distance_threshold_col is None:
        distance_threshold = None
    else:
        # \vec eta = (\eta_row, \eta_col)
        distance_threshold = (distance_threshold_row, distance_threshold_col)

    # distance_threshold = None if (distance_threshold_row is None or distance_threshold_col is None) else (distance_threshold_row, distance_threshold_col) #\vec eta = (\eta_row, \eta_col)
    return NearestNeighborImputer(estimator, data_type, distance_threshold)



import numpy as np
from typing import Callable
import numpy.typing as npt
import pandas as pd
from nsquared.datasets.dataloader_base import NNDataLoader
from nsquared.data_types import DataType


def empirical_quantile_function(samples: npt.NDArray) -> Callable:
    """Create an empirical quantile function from the samples.

    Args:
        samples (npt.NDArray): Input samples for quantile function.

    Returns:
        Callable: Quantile function that takes a single argument and returns quantile values.

    """
    samples_diff = np.concatenate([np.array(samples[0]).reshape(1), np.diff(samples)])

    def quantile_function(q: npt.NDArray) -> npt.NDArray:
        """Compute quantile values based on the input quantile.

        Args:
            q (npt.NDArray): Quantile values to compute.

        Returns:
            npt.NDArray: Computed quantile values.

        """
        # Compute the empirical CDF values
        n = len(samples)
        cdf = np.arange(1, n + 1) / n
        # Use broadcasting to calculate the Heaviside contributions
        heaviside_matrix = np.heaviside(
            np.expand_dims(q, 1) - np.expand_dims(cdf, 0), 0.0
        )
        # Add a column of ones to the left of the Heaviside matrix
        first_col = np.ones(heaviside_matrix.shape[0]).reshape(-1, 1)
        heaviside_matrix = np.concatenate([first_col, heaviside_matrix], axis=1)
        # Remove the last column of Heaviside_matrix
        heaviside_matrix = heaviside_matrix[:, :-1]
        # Compute quantile values by summing contributions
        quantile_values = heaviside_matrix @ samples_diff

        return quantile_values

    return quantile_function


class EarningsDataLoader(NNDataLoader):
    """Data loader for earnings dataset."""

    def __init__(
        self,
        quarterly_data: dict,
        quarterly_actual: dict,
        ibes_tickers: list[str],
        current_time: pd.Timestamp,
        agg: str = "mean",
        save_processed: bool = False,
    ):
        """Initializes the EarningsDataLoader.

        Args:
            quarterly_data (dict): Quarterly earnings data.
            quarterly_actual (dict): Actual earnings data.
            ibes_tickers (list[str]): List of IBES tickers.
            current_time (pd.Timestamp): Current time for filtering data.
            agg (str, optional): Aggregation method. Defaults to "mean".
            save_processed (bool, optional): Whether to save processed data. Defaults to False.

        """
        super().__init__(agg=agg, save_processed=save_processed)
        self.quarterly_actual = quarterly_actual
        self.ibes_tickers = ibes_tickers
        self.current_time = current_time
        self.rows = [oftic for oftic, _ in self.ibes_tickers]
        self.cols = [
            (year, quarter) for year in range(2010, 2025) for quarter in range(1, 5)
        ]

        self.supported_aggs = ["mean", "sum", "median", "std", "variance"]
        if self.agg == "mean":
            self.agg_func = np.mean
        elif self.agg == "sum":
            self.agg_func = np.sum
        elif self.agg == "median":
            self.agg_func = np.median
        elif self.agg == "std":
            self.agg_func = np.std
        elif self.agg == "variance":
            self.agg_func = np.var
        else:
            raise ValueError(
                f"Aggregation method {self.agg} not supported. Supported methods: {self.supported_aggs}"
            )

        # Filter the data to the current time
        self.quarterly_data = dict()
        for k, v in quarterly_data.items():
            if v is None:
                self.quarterly_data[k] = None
                continue
            self.quarterly_data[k] = v[v["ann_datetime"] <= self.current_time]
            if self.quarterly_data[k].shape[0] == 0:
                self.quarterly_data[k] = None

    def process_data_scalar(self) -> tuple[npt.NDArray, npt.NDArray]:
        """Processes the data into scalar matrix setting.

        Returns:
            tuple[npt.NDArray, npt.NDArray]: Tuple of 2D data matrix and mask.

        """
        data_matrix = np.zeros((len(self.rows), len(self.cols)))
        mask_matrix = np.ones((len(self.rows), len(self.cols)))

        for i, oftic in enumerate(self.rows):
            for j, (year, quarter) in enumerate(self.cols):
                data = self.quarterly_data.get((oftic, year, quarter))
                if data is not None:
                    data_matrix[i, j] = self.agg_func(data["value"])
                    mask_matrix[i, j] = 1
                else:
                    data_matrix[i, j] = np.nan
                    mask_matrix[i, j] = 0

        return data_matrix, mask_matrix

    def process_data_distribution(
        self, data_type: DataType | None = None
    ) -> tuple[npt.NDArray, npt.NDArray]:
        """Processes the data into distributional setting.

        Args:
            data_type: Data type to process. Default: None.

        Returns:
            tuple[npt.NDArray, npt.NDArray]: Tuple of 2D data matrix and mask.

        """
        data_matrix = np.empty((len(self.rows), len(self.cols)), dtype=object)
        mask_matrix = np.ones((len(self.rows), len(self.cols)))

        for i, oftic in enumerate(self.rows):
            for j, (year, quarter) in enumerate(self.cols):
                data = self.quarterly_data.get((oftic, year, quarter))
                if data is not None:
                    data_matrix[i, j] = empirical_quantile_function(
                        data["value"].to_numpy()
                    )
                    mask_matrix[i, j] = 1
                else:
                    data_matrix[i, j] = np.nan
                    mask_matrix[i, j] = 0

        return data_matrix, mask_matrix

    def get_full_state_as_dict(self, include_metadata: bool = False) -> dict:
        """Returns the full state of the data loader as a dictionary.

        Args:
            include_metadata (bool, optional): Whether to include metadata in the dictionary. Defaults to False.

        Returns:
            dict: Full state of the data loader.

        """
        state = {
            "quarterly_data": self.quarterly_data,
            "quarterly_actual": self.quarterly_actual,
            "ibes_tickers": self.ibes_tickers,
            "current_time": self.current_time,
        }
        if include_metadata:
            state["metadata"] = {"agg": self.agg, "save_processed": self.save_processed}
        return state



from typing import Dict, Type
from importlib import import_module
from .dataloader_base import NNDataLoader
from typing import Any, Tuple

# stores available dataset loaders
_DATASETS: Dict[str, Type[NNDataLoader]] = {}
_DATASET_PARAMS: Dict[str, Dict[str, Tuple[Type, Any, str]]] = {}


def register_dataset(name: str, params: Dict[str, Tuple[Type, Any, str]] = {}) -> Any:
    """Decorator to register a dataset loader."""

    def decorator(cls: Type[NNDataLoader]) -> Type[NNDataLoader]:
        _DATASETS[name] = cls
        if params:
            _DATASET_PARAMS[name] = params
        return cls

    return decorator


def get_available_datasets() -> list[str]:
    """Returns the available dataset loaders."""
    return list(_DATASETS.keys())


class NNData:
    """Factory class to create dataset instances."""

    @staticmethod
    def create(
        dataset_name: str, download: bool = False, save_dir: str = "./", **kwargs: Any
    ) -> NNDataLoader:
        """Create a dataset loader instance by name.

        Args:
            dataset_name: Name of the dataset
            download: Whether to download the data locally. Default: False. If True, data is downloaded at save_dir
            save_dir: Directory to save the data. Default: "./" (current directory).
            **kwargs: Additional arguments to be passed to the dataset loader.

        Returns:
            An instance of the requested dataset loader

        Raises:
            ValueError: If the dataset name is not registered

        """
        if dataset_name not in _DATASETS:
            try:
                import_module(f"nsquared.datasets.{dataset_name}")
            except ImportError:
                raise ValueError(
                    f"Dataset {dataset_name} not found. Available datasets: {get_available_datasets()}"
                )

        return _DATASETS[dataset_name](download=download, save_dir=save_dir, **kwargs)

    @staticmethod
    def get_data_params(dataset_name: str) -> Dict[str, Tuple[Type, Any, str]]:
        """Get the custom parameters required for a dataset loader.

        Args:
            dataset_name: Name of the dataset

        """
        if dataset_name not in _DATASET_PARAMS:
            try:
                import_module(f"nsquared.datasets.{dataset_name}")
            except ImportError:
                raise ValueError(
                    f"Dataset {dataset_name} not found. Available datasets: {get_available_datasets()}"
                )
        return _DATASET_PARAMS.get(dataset_name, {})

    @staticmethod
    def help(dataset_name: str = "") -> None:
        """Display help information for datasets.

        Args:
            dataset_name: Optional name of dataset to get specific help for.
                          If "", lists all available datasets.

        """
        if dataset_name == "":
            print("Available datasets:")
            for name in get_available_datasets():
                print(f"  - {name}")
            print(
                "\nUse NNData.help('dataset_name') to get help for a specific dataset."
            )
            return

        if dataset_name not in _DATASETS:
            try:
                import_module(f"nsquared.datasets.{dataset_name}")
            except ImportError:
                print(
                    f"Dataset {dataset_name} not found. Available datasets: {get_available_datasets()}"
                )
                return

        # Get the dataset class
        dataset_class = _DATASETS[dataset_name]

        # Print general information
        print(f"Help for dataset: {dataset_name}")
        print("=" * 50)
        print(dataset_class.__doc__)

        # Print common parameters
        print("\nCommon parameters:")
        print("  download: bool = False")
        print("      Whether to download the data locally")
        print("  save_dir: str = './'")
        print("      Directory to download and save data")
        print("  agg: str = 'mean'")
        print(
            "      Aggregation method (options: 'mean', 'sum', 'median', 'std', 'variance')"
        )
        print("  save_processed: bool = False")
        print("      Whether to save processed data to disk")

        # Print dataset-specific parameters from registry
        params = NNData.get_data_params(dataset_name)
        if params:
            print("\nDataset-specific parameters:")
            for name, (param_type, default, description) in params.items():
                type_name = param_type.__name__
                print(f"  {name}: {type_name} = {default}")
                print(f"      {description}")



from fancyimpute import SoftImpute, BiScaler
import numpy.typing as npt


def softimpute(X: npt.NDArray) -> npt.NDArray:
    """SoftImpute imputation method.

    Parameters
    ----------
    X : npt.NDArray
        N x T input data matrix with missing values as np.nan.

    Returns
    -------
    X_imputed : npt.NDArray
        The imputed data matrix.

    """
    # Create a SoftImpute instance with the provided arguments
    softimpute = SoftImpute(normalizer=BiScaler(), verbose=False)
    return softimpute.fit_transform(X)



"""Implementation of vanilla NN imputer.

Details of the vanilla NN algorithm can be found in Sec. 2.3 of the following paper:
    Dwivedi, R., Tian, K., Tomkins, S., Klasnja, P., Murphy, S., & Shah, D. (2022).
    Counterfactual inference for sequential experiments. arXiv preprint arXiv:2202.06891.
"""

from .nnimputer import NearestNeighborImputer
from .data_types import Scalar
from .estimation_methods import RowRowEstimator, ColColEstimator
from typing import Optional


def row_row(
    distance_threshold: Optional[float] = None, is_percentile: bool = True
) -> NearestNeighborImputer:
    """Create a row-row nearest neighbor imputer.

    Args:
        distance_threshold (float, optional): Distance threshold for nearest neighbors. Defaults to None.
        is_percentile (bool, optional): Whether to use percentile-based distance threshold. Defaults to True.

    Returns:
        NearestNeighborImputer: A row-row nearest neighbor imputer.

    """
    estimator = RowRowEstimator(is_percentile=is_percentile)
    data_type = Scalar()
    return NearestNeighborImputer(estimator, data_type, distance_threshold)


def col_col(
    distance_threshold: Optional[float] = None, is_percentile: bool = True
) -> NearestNeighborImputer:
    """Create a column-column nearest neighbor imputer.

    Args:
        distance_threshold (float, optional): Distance threshold for nearest neighbors. Defaults to None.
        is_percentile (bool, optional): Whether to use percentile-based distance threshold. Defaults to True.

    Returns:
        NearestNeighborImputer: A column-column nearest neighbor imputer.

    """
    estimator = ColColEstimator(is_percentile=is_percentile)
    data_type = Scalar()
    return NearestNeighborImputer(estimator, data_type, distance_threshold)



import subprocess
import pytest


def test_pre_commit() -> None:
    """Test that pre-commit hooks run successfully."""
    result = subprocess.run(
        ["pre-commit", "run", "--all-files"], capture_output=True, text=True
    )
    assert result.returncode == 0, (
        f"Pre-commit hooks failed:\n{result.stdout}\n{result.stderr}"
    )


if __name__ == "__main__":
    pytest.main()



import numpy as np
from nsquared.dr_nn import dr_nn
import pytest

ROWS = 10
COLS = 10


def test_drnn_all0() -> None:
    """Test the drnn imputer with all values 0."""
    imputer = dr_nn(distance_threshold_row=1, distance_threshold_col=1)
    data = np.zeros([ROWS, COLS])
    mask = np.ones([ROWS, COLS], dtype=bool)

    imputed_value = imputer.impute(
        row=0,
        column=0,
        data_array=data,
        mask_array=mask,
    )

    # imputed value should be 0
    assert imputed_value == 0


def test_drnn_all1() -> None:
    """Test the drnn imputer with all values 1."""
    imputer = dr_nn(distance_threshold_row=1, distance_threshold_col=1)
    data = np.ones([ROWS, COLS])
    mask = np.ones([ROWS, COLS], dtype=bool)

    imputed_value = imputer.impute(
        row=0,
        column=0,
        data_array=data,
        mask_array=mask,
    )

    # imputed value should be 1
    assert imputed_value == 1


def test_drnn_avg1() -> None:
    """Test the drnn imputer with single doubly robust estimate to average over."""
    imputer = dr_nn(distance_threshold_row=1, distance_threshold_col=1)
    data = np.array([[np.nan, 1, 4.0], [1, 1, 3.0], [1, 5, 2]])

    mask = np.array(
        [
            [0, 1, 0],  # Missing value at [0, 0]
            [1, 1, 0],
            [0, 0, 0],
        ]
    )

    # Manually call the imputation logic for the missing value at [0, 0]
    imputed_value = imputer.impute(
        row=0,
        column=0,
        data_array=data,
        mask_array=mask,
    )

    # The imputed value should be 1 + 1 - 1 = 1
    assert imputed_value == 1


def test_drnn_avg2() -> None:
    """Test the drnn imputer with multiple doubly robust estimates to average over."""
    imputer = dr_nn(distance_threshold_row=0.5, distance_threshold_col=0.5)
    data = np.array([[0, 1, 0, 1], [1, 1, 0, 1], [0, 0, 1, 0], [1, 1, 0, 1]])
    # neighbors are row 1, row 3, col 1, col 3. So estimate is:
    # ((1 + 1 - 1) + (1 + 1 - 1) + (1 + 1 - 1) + (1 + 1 - 1)) / 4 = 1

    mask = np.array([[0, 1, 1, 1], [1, 1, 1, 1], [1, 1, 1, 1], [1, 1, 1, 1]])

    imputed_value = imputer.impute(
        row=0,
        column=0,
        data_array=data,
        mask_array=mask,
    )
    assert imputed_value == 1


if __name__ == "__main__":
    pytest.main()



"""A function to return the DRNN imputer.

More details on the Doubly Robust Nearest Neighbor (DRNN) method can be found in Section 2.2.2 of the following paper:
    Dwivedi, R., Tian, K., Tomkins, S., Klasnja, P., Murphy, S., & Shah, D. (2022).
    Doubly robust nearest neighbors in factor models. arXiv preprint arXiv:2211.14297.
"""

from .nnimputer import NearestNeighborImputer
from typing import Optional


def dr_nn(
    distance_threshold_row: Optional[float] = None,
    distance_threshold_col: Optional[float] = None,
    is_percentile: bool = True,
) -> NearestNeighborImputer:
    """Create a doubly robust nearest neighbor imputer.

    If distance_threshold_row and distance_threshold_col are not provided, they must be set by calling fit on the imputer.


    Args:
        distance_threshold_row (float): [Optional] Distance threshold for row-row nearest neighbors.
        distance_threshold_col (float): [Optional] Distance threshold for column-column nearest neighbors.
        is_percentile (bool): [Optional] Whether to use percentile-based distance threshold. Defaults to True.

    Returns:
        NearestNeighborImputer: A doubly robust nearest neighbor imputer.

    """
    from .estimation_methods import DREstimator
    from .data_types import Scalar

    estimator = DREstimator(is_percentile=is_percentile)
    data_type = Scalar()
    # note that the default value of distance_threshold is np.inf -> distance threshold is unused for DRNN
    if distance_threshold_row is None or distance_threshold_col is None:
        distance_threshold = None
    else:
        distance_threshold = (distance_threshold_row, distance_threshold_col)
    return NearestNeighborImputer(estimator, data_type, distance_threshold)



"""A function to return the Adaptively-weighted Nearest Neighbors imputer."""

from typing import Optional

# from nsquared.weighted_estimation_methods import AWNNEstimator
from nsquared.data_types import Scalar
from nsquared.nnimputer import NearestNeighborImputer
from nsquared.estimation_methods import AWNNEstimator


def aw_nn(
    distance_threshold: Optional[float] = 0,
    delta: float = 1,
    convergence_threshold: float = 1e-4,
    max_iterations: int = 10,
    noise_variance: Optional[float] = None,
) -> NearestNeighborImputer:
    """Create a AWNN imputer."""
    # distance_threshold defaults to 0 and is wholly ignored by the AWNN approach.
    estimator = AWNNEstimator(
        delta=delta,
        convergence_threshold=convergence_threshold,
        max_iterations=max_iterations,
        noise_variance=noise_variance,
    )
    data_type = Scalar()
    return NearestNeighborImputer(estimator, data_type, distance_threshold)



"""Dataset loader for the MovieLens 1M dataset.

Source: https://grouplens.org/datasets/movielens/1m/
License: CC-BY-4.0
Paper Reference for transformation implementation:
    Yihua Li, Devavrat Shah, Dogyoon Song, and Christina Lee Yu.
    "Nearest Neighbors for Matrix Estimation Interpreted as Blind Regression for Latent Variable Model."
    IEEE Transactions on Information Theory, 2020.
    https://doi.org/10.1109/TIT.2019.2950299
"""

from nsquared.datasets.dataloader_base import NNDataLoader
from nsquared.datasets.dataloader_factory import register_dataset
from nsquared.data_types import DataType

import numpy as np
import pandas as pd
from typing import Any
import logging
from joblib import Memory
import zipfile
import os
import requests
import shutil

memory = Memory(".joblib_cache", verbose=2)

logger = logging.getLogger(__name__)

params = {
    "sample_users": (
        int,
        None,
        "Number of users to sample from the dataset. By default, returns all.",
    ),
    "sample_movies": (
        int,
        None,
        "Number of movies to sample from the dataset. By default, returns all.",
    ),
    "seed": (int, None, "Random seed for reproducibility"),
}


@register_dataset("movielens", params)
class MovieLensDataLoader(NNDataLoader):
    """Data from the MovieLens study formatted into a matrix or tensor.
    To initialize with default settings, use: NNData.create("movielens").

    """

    urls = {
        "full_archive": "https://files.grouplens.org/datasets/movielens/ml-1m.zip",
    }

    def __init__(
        self,
        sample_users: int | None = None,
        sample_movies: int | None = None,
        seed: int | None = None,
        **kwargs: Any,
    ):
        """Initializes the MovieLens data loader.

        Args:
        ----
            sample_users: Number of users to sample from the dataset. Default: None (use all users).
            sample_movies: Number of movies to sample from the dataset. Default: None (use all movies).
            seed: Random seed for reproducibility. Default: None
            kwargs: Additional keyword arguments.

        """
        super().__init__(
            **kwargs,
        )
        self.sample_users = sample_users
        self.sample_movies = sample_movies
        if seed is not None:
            np.random.seed(
                seed=seed
            )  # instantiate random seed if provided but do it only once here

    def process_data_scalar(self, agg: str = "mean") -> tuple[np.ndarray, np.ndarray]:
        """Process the data into scalar setting. Note that this implementation is specific to MovieLens as it calls upon functions that do specific MovieLens data processing.

        Args:
        ----
            agg: Aggregation method to use. Default: "mean". Options: "mean", "sum", "median", "std", "variance"

        Returns:
        -------
            data: 2d processed data matrix of floats
            mask: Mask for processed data

        """
        df_movies, df_ratings, df_users = self._load_data()
        data_df = df_ratings.pivot(
            index="UserID", columns="MovieID", values="Rating"
        )  # rows: UserID, columns: MovieID, values: Rating
        if self.sample_users is not None:
            data_df = data_df.sample(n=self.sample_users)
        if self.sample_movies is not None:
            data_df = data_df.sample(n=self.sample_movies, axis=1)
        mask_df = data_df.notna().astype(bool)
        data = data_df.to_numpy()
        mask = mask_df.to_numpy()
        self.data = data
        self.mask = mask
        return data, mask

    def process_data_distribution(
        self, data_type: DataType | None = None
    ) -> tuple[np.ndarray, np.ndarray]:
        """Does not apply to MovieLens as it is not a distributional dataset.

        Args:
            data_type: Data type to process. Default: None.

        """
        raise (ValueError("There is no distributional data for MovieLens."))

    def get_full_state_as_dict(self, include_metadata: bool = False) -> dict:
        """Returns the full state as a dictionary. For HeartSteps, this includes the data, masking matrix, and the custom parameters (if include_metadata == True

        If the data and mask are None, then the data has not been processed yet. Call process_data_scalar() or process_data_distribution() to process the data first.

        Args:
            include_metadata (bool): Whether to include metadata in the dictionary. Default: False. The metadata for HeartSteps is currently empty.

        """
        full_state = {
            "data": self.data,
            "mask": self.mask,
            "custom_params": {
                "sample_users": self.sample_users,
                "sample_movies": self.sample_movies,
            },
        }
        return full_state

    @classmethod
    @memory.cache
    def _load_data(cls) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """Download and extract the MovieLens 1M dataset and load its contents."""
        zip_path = "movielens_1m.zip"
        extract_dir = "movielens_1m"

        # Download if missing or previously corrupted
        needs_download = not os.path.exists(zip_path)
        if not needs_download:
            try:
                with zipfile.ZipFile(zip_path, "r") as test_zip:
                    test_zip.testzip()
            except zipfile.BadZipFile:
                os.remove(zip_path)
                needs_download = True

        if needs_download:
            logger.info("Downloading MovieLens 1M dataset...")
            response = requests.get(cls.urls["full_archive"], stream=True)
            if response.status_code != 200:
                raise RuntimeError(
                    f"Failed to download dataset: {response.status_code}"
                )
            with open(zip_path, "wb") as f:
                shutil.copyfileobj(response.raw, f)

        if not os.path.exists(extract_dir):
            logger.info("Extracting dataset...")
            with zipfile.ZipFile(zip_path, "r") as zip_ref:
                zip_ref.extractall(extract_dir)

        base = os.path.join(extract_dir, "ml-1m")
        df_movies = pd.read_csv(
            os.path.join(base, "movies.dat"),
            sep="::",
            header=None,
            engine="python",
            encoding="ISO-8859-1",
            names=["MovieID", "Title", "Genres"],
        )
        df_ratings = pd.read_csv(
            os.path.join(base, "ratings.dat"),
            sep="::",
            header=None,
            engine="python",
            encoding="ISO-8859-1",
            names=["UserID", "MovieID", "Rating", "Timestamp"],
        )
        df_users = pd.read_csv(
            os.path.join(base, "users.dat"),
            sep="::",
            header=None,
            engine="python",
            encoding="ISO-8859-1",
            names=["UserID", "Gender", "Age", "Occupation", "Zip-code"],
        )
        # delete the zip file
        if os.path.exists(zip_path):
            os.remove(zip_path)
        return df_movies, df_ratings, df_users



"""Implementation of Nadaraya-Watson estimation method.

Instead of identifying the nearest neighbors using a thresholding rule
(i.e., neighbors within a certain distance), the NW estimator uses all
neighbors, but weights their contribution according to a kernel function.

TODO (Albert): Test NW on HeartSteps dataset with various kernel functions
TODO (Albert): Implement additional kernel functions (e.g., Epanechnikov)
"""

from .utils.kernels import gaussian, laplace, singular_box, box
from .nnimputer import EstimationMethod, DataType
from typing import Union, Tuple, Any

import numpy.typing as npt
import numpy as np


class NadarayaWatsonEstimator(EstimationMethod):
    valid_kernels = ["gaussian", "laplace", "singular_box", "box"]

    def __init__(self, kernel: str = "gaussian"):
        """Initialize the Nadaraya-Watson estimator.

        Args:
            kernel (str, optional): The kernel to use. Defaults to "gaussian".

        Raises:
            ValueError: If the kernel is not valid.

        """
        if self.kernel not in self.valid_kernels:
            raise ValueError(
                f"{self.kernel=} is not a valid kernel. Currently supported kernels are {', '.join(self.valid_kernels)}"
            )
        self.kernel = kernel

        super().__init__()

    def __str__(self):
        return f"NadarayaWatsonEstimator(kernel={self.kernel})"

    def impute(
        self,
        row: int,
        column: int,
        data_array: npt.NDArray,
        mask_array: npt.NDArray,
        distance_threshold: Union[float, Tuple[float, float]],
        data_type: DataType,
        allow_self_neighbor: bool = False,
        **kwargs: Any,
    ) -> npt.NDArray:
        """Impute the missing value at the given row and column.

        Parameters
        ----------
        row : int
            The row index of the missing value.
        column : int
            The column index of the missing value.
        data_array : npt.NDArray
            The data array.
        mask_array : npt.NDArray
            The mask array.
        distance_threshold : float
            The distance threshold.
        data_type : DataType
            The data type.
        allow_self_neighbor : bool, optional
            Whether to allow the entry itself as a neighbor, by default False
        **kwargs : Any
            Additional keyword arguments

        Returns
        -------
        npt.NDArray
            The imputed value.

        """
        if isinstance(distance_threshold, tuple):
            raise ValueError(
                "The Nadaraya-Watson estimator only accepts a single distance threshold."
            )

        data_shape = data_array.shape
        n_rows = data_shape[0]
        n_cols = data_shape[1]

        # Calculate distances between rows
        row_distances = np.zeros(n_rows)
        for i in range(n_rows):
            # Get columns observed in both row i and row
            overlap_columns = np.logical_and(mask_array[row], mask_array[i])

            if not np.any(overlap_columns):
                row_distances[i] = np.inf
                continue

            # Calculate distance between rows
            for j in range(n_cols):
                if (
                    not overlap_columns[j] or j == column
                ):  # Skip missing values and the target column
                    continue
                row_distances[i] += data_type.distance(
                    data_array[row, j], data_array[i, j]
                )
                row_distances[i] /= np.sum(overlap_columns)

        match self.kernel:
            case "gaussian":
                K = gaussian(dists=row_distances, eta=distance_threshold)
            case "laplace":
                K = laplace(dists=row_distances, eta=distance_threshold)
            case "singular_box":
                K = singular_box(dists=row_distances, eta=distance_threshold)
            case "box":
                K = box(dists=row_distances, eta=distance_threshold)
            case _:
                raise ValueError(f"{self.kernel=} is not supported")

        assert K.shape == (n_rows,)
        K = K.reshape(-1, 1)

        K_sum = K.sum(axis=0)
        K_sum_nan = np.where(K_sum == 0, np.nan, K_sum)
        # set y to be the column-th column of data_array
        y = data_array[:, column]
        assert y.shape == (n_rows,)
        pred = y @ K / K_sum_nan

        # get index of nan in K along axis=0
        nan_idx = np.argwhere(np.isnan(K))
        for i, j in nan_idx:
            # NOTE: for singular kernels, anytime the row-th object is one of the
            # objects in the training set (indicated by nan values of K),
            # we set the corresponding entry of pred to the actual value of y
            #  at the row-th object
            pred[j] = y[i]
        # if an entry of K_sum is zero, set the corresponding entry of pred to zero
        pred = np.where(K_sum == 0, 0, pred)

        return pred
